from scipy.ndimage import median_filter
from ..segment import prediction2foreground, prediction2seed
from ..utils import *
from skimage import filters, morphology, measure, feature, segmentation
import pandas as pd
import numpy as np
from matplotlib import pyplot as plt


class CellTracker:

    def __init__(self,
                 sorted_patches,  # load patches patches as list
                 time_points,
                 pixel_classifier,
                 seed_prob_min=0.2,
                 edge_prob_max=0.9,
                 background_prob_max=0.2,
                 min_overlap_threshold=0.5,
                 min_overlap_area=100,
                 min_size_similarity=0.75,
                 min_cell_size=200,
                 min_seed_size=10,
                 max_iter=5,
                 cache_image_features=False,
                 verbose=True,
                 backtrack_generations=4,
                 no_merge=False,
                 no_split=False,
                 hard_split=False):

        if len(time_points) != len(sorted_patches):
            raise ValueError("The number of time points given doesn't match the number of image patches!")
        self.timepoints = time_points
        self.frames = sorted_patches
        self.pixel_classifier = pixel_classifier
        self.seed_prob_min = seed_prob_min
        self.edge_prob_max = edge_prob_max
        self.background_prob_max = background_prob_max
        self.min_seed_size = min_seed_size
        self.cache_in = cache_image_features

        self.min_overlap_threshold = min_overlap_threshold
        self.min_overlap_area = min_overlap_area
        self.min_size_similarity = min_size_similarity
        self.min_cell_size = min_cell_size
        self.backtrack_generations = backtrack_generations
        self.max_iter = max_iter

        self.regionprops = None
        self.overlap_stats = None
        self.to_be_merged = [-1]
        self.to_be_split = [-1]
        self.verbose = verbose
        self.no_merge = no_merge
        self.no_split = no_split
        self.hard_split = hard_split

    def locate_spurious_events(self):
        to_be_split, to_be_merged = _locate_spurious_events(self.regionprops, verbose=self.verbose)
        self.to_be_split = to_be_split
        self.to_be_merged = to_be_merged

    def update_masks(self):
        if len(self.to_be_merged) > 0:
            if not self.no_merge:
                for c in self.to_be_merged:
                    self._merge_cells(c)
        if len(self.to_be_split) > 0:
            if not self.no_split:
                for c in self.to_be_split:
                    self._split_cells(c)

    def update_frames(self):
        _regionprops = []
        for i, p in enumerate(self.frames):
            p.locate_particles()
            p.regionprops['$cell'] = ['{}_{}'.format(i, l) for l in p.regionprops.index]
            p.regionprops['$time'] = i
            _regionprops.append(p.regionprops)
        self.regionprops = pd.concat(_regionprops).set_index('$cell')

    def _update_regionprops(self):
        _regionprops = []
        for i, p in enumerate(self.frames):
            _regionprops.append(p.regionprops)
            p.regionprops['$cell'] = ['{}_{}'.format(i, l) for l in p.regionprops.index]
        _regionprops = pd.concat(_regionprops).set_index('$cell')
        for col in _regionprops.columns:
            if col not in self.regionprops.columns:
                self.regionprops[col] = _regionprops.loc[self.regionprops.index][col]
        del _regionprops

    def trace_by_overlap(self):
        counter = 0
        while counter <= self.max_iter:
            self.update_frames()
            self.link_cells_by_overlap()
            self.locate_spurious_events()
            self.update_masks()
            if self._end_iter():
                self.update_frames()
                self.link_cells_by_overlap()
                break
            counter += 1

    def link_cells_by_overlap(self):
        self.regionprops, self.overlap_stats = overlap_match(self.frames,
                                                             self.regionprops,
                                                             min_cell_size=self.min_cell_size,
                                                             min_size_similarity=self.min_size_similarity,
                                                             min_overlap_threshold=self.min_overlap_threshold,
                                                             min_overlap_area=self.min_overlap_area)

    def _end_iter(self):
        end = False
        if (len(self.to_be_merged) + len(self.to_be_split)) == 0:
            end = True
        return end

    def _merge_cells(self, cell_list):

        coords = self.regionprops.loc[np.array(cell_list)]['$coords']
        dilated_coords = [dilate_by_coords(x, self.frames[0].shape, morphology.disk(1)) for x in coords]
        unique_coords, pix_count = np.unique(np.vstack(dilated_coords), return_counts=True, axis=0)
        border_coords = unique_coords[pix_count >= 2]
        merged_coords = np.vstack(list(coords) + [border_coords])
        time = np.unique([int(x.split('_')[0]) for x in cell_list])[0]
        new_label = np.min([int(x.split('_')[1]) for x in cell_list])
        self.frames[time].labeled_mask[merged_coords[:, 0], merged_coords[:, 1]] = new_label

    def _split_cells(self, cell_ref):
        cell, n_daughter = cell_ref
        time, label = np.array(cell.split('_')).astype(int)
        x, y = self.regionprops.loc[cell]['$coords'].T
        cropped_prob = np.zeros(self.frames[time].prob_mask.shape)
        cropped_mask = np.zeros(self.frames[time].mask.shape)
        cropped_mask[x, y] = 1
        cropped_prob[x, y, :] = self.frames[time].prob_mask[x, y, :]

        for th in np.linspace(self.seed_prob_min, 1, 5):
            seed = prediction2seed(cropped_prob,
                                   seed_min=th,
                                   edge_max=self.edge_prob_max,
                                   min_seed_size=self.min_seed_size)

            if seed.max() >= n_daughter:
                break

        if seed.max() != n_daughter and self.hard_split:
            seed = np.zeros(self.frames[time].mask.shape)
            current_mask = self.frames[time].labeled_mask == label
            prev_frame = (self.frames[time - 1].labeled_mask) * current_mask
            counter = 0
            for d in np.unique(prev_frame):
                if d != 0:
                    if (np.sum(prev_frame == d) / len(x)) > 0.1:
                        seed[prev_frame == d] = counter + 1
                        counter += 1
        if seed.max() == n_daughter:
            seed[seed == 1] = self.frames[time].labeled_mask.max() + 1
            seed[seed == 2] = self.frames[time].labeled_mask.max() + 2
            cropped_watershed = segmentation.watershed(image=cropped_prob[:, :, 1],
                                                       mask=cropped_mask,
                                                       markers=seed,
                                                       connectivity=1,
                                                       compactness=0.1,
                                                       watershed_line=False)
            new_labeled_mask = self.frames[time].labeled_mask.copy()
            new_labeled_mask[x, y] = cropped_watershed[x, y]
            self.frames[time].labeled_mask = new_labeled_mask

    def plot_cell(self, cell_id):
        show_highlight(cell_id, self.frames)

    def trace_lineage(self):
        rev_lineage = {}
        init_cell_counter = 1
        for c in self.regionprops.index:
            daughters = self.regionprops.loc[c]['daughter(s)']
            if c not in rev_lineage:
                rev_lineage[c] = str(init_cell_counter)
                init_cell_counter += 1
            if len(daughters) == 1:
                rev_lineage[daughters[0]] = rev_lineage[c]
            else:
                for i, d in enumerate(daughters):
                    rev_lineage[d] = '{}.{}'.format(rev_lineage[c], i + 1)
        lineage = [rev_lineage[c] for c in self.regionprops.index]
        self.regionprops['cell_lineage'] = lineage

    def refine_trace(self):
        _refine_trajectories(self,
                             seed_threshold=self.seed_prob_min,
                             force_split=self.hard_split,
                             trace_generations=self.backtrack_generations)

    def link_cells(self):
        _link_cells(self)


def _refine_trajectories(tracker,
                         seed_threshold=0.7,
                         force_split=True,
                         trace_generations=3):
    from skimage import measure, segmentation, morphology
    to_be_split, to_be_merged = _locate_spurious_events(tracker.regionprops, verbose=False)
    new_labels = [f.labeled_mask.copy() for f in tracker.frames]  # copy mask
    from scipy.stats import mode
    import warnings
    warnings.filterwarnings("ignore")
    for pair in to_be_merged:
        t, l1 = pair[0].split('_')
        t, l2 = pair[1].split('_')
        t = int(t)
        coords1 = tracker.regionprops.loc[pair[0], '$coords']
        coords2 = tracker.regionprops.loc[pair[1], '$coords']
        x, y = np.vstack([coords1, coords2]).T
        max_label = new_labels[t].max()
        new_labels[t][x, y] = max_label + 1

    for cell in to_be_split:
        t, l1 = cell.split('_')
        t = int(t)
        prob = tracker.frames[t].prob_mask[:, :, 2]
        mask = tracker.frames[t].labeled_mask == int(l1)
        eroded_mask = morphology.erosion(mask)
        masked_prob = prob * eroded_mask
        sink = 1 - prob
        seed_found = False
        for th in np.linspace(seed_threshold, 1, 10):
            seeds = masked_prob > th
            labeled_seeds = measure.label(seeds)
            if labeled_seeds.max() == 2:
                seed_found = True
                break
        if not seed_found and force_split:
            parents = back_track(tracker.regionprops, cell, n_generations=trace_generations)
            daughters = forward_track(tracker.regionprops, cell, n_generations=trace_generations)
            nearest_split_event = [np.inf, []]
            for p in parents:
                if len(p) == 2:
                    _t = int(p[0].split('_')[0])
                    dt = np.abs(_t - t)
                    if dt < nearest_split_event[0]:
                        nearest_split_event = [_t, p]
                    break
            for p in daughters:
                if len(p) == 2:
                    _t = int(p[0].split('_')[0])
                    dt = np.abs(_t - t)
                    if dt < nearest_split_event[0]:
                        nearest_split_event = [_t, p]
                    break
            seeds = np.zeros(mask.shape)
            if len(nearest_split_event[1]) > 0:
                x1, y1 = tracker.regionprops.loc[nearest_split_event[1][0], '$coords'].T
                x2, y2 = tracker.regionprops.loc[nearest_split_event[1][1], '$coords'].T
                seeds[x1, y1] = 1
                seeds[x2, y2] = 2
                labeled_seeds = measure.label(seeds) * mask
                seed_found = True
        if seed_found:
            watersheded = segmentation.watershed(image=sink, markers=labeled_seeds, mask=mask, compactness=100)
            for i in np.unique(watersheded):
                max_label = new_labels[t].max()
                if i > 0:
                    new_labels[t][watersheded == i] = max_label + i
        else:
            print('Rectification failed for cell {}.'.format(cell))
    for i, p in enumerate(tracker.frames):
        # remove tiny chunks that don't belong
        labeled_mask = morphology.remove_small_objects(new_labels[i].copy(), tracker.min_seed_size)
        p.labeled_mask = labeled_mask
        p.regionprops = None
        p.locate_particles(precompute_contours=False)


def back_track(regionprops,
               cell_id,
               n_generations=3):
    # the backtrack algorithm is based on the naive hypothesis that the fragments of an oversegmented cell
    # should have a common ancester tracing back n generations whereas an undersegmented cell should not.
    from scipy.stats import mode
    import warnings
    warnings.filterwarnings("ignore")
    time, label = np.array(cell_id.split('_')).astype(int)
    elderlist = []
    current_cell = np.array([cell_id])
    while True:
        if n_generations == 0 or time == 0:
            break
        mothers = np.unique(np.hstack(regionprops.loc[current_cell]['mother(s)'].values))
        current_cell = mothers
        n_generations -= 1
        time -= 1
        if len(mothers) > 0:
            elderlist.append(mothers)
        else:
            break
    return elderlist


def forward_track(regionprops,
                  cell_id,
                  n_generations=3):
    # the backtrack algorithm is based on the naive hypothesis that the fragments of an oversegmented cell
    # should have a common ancester tracing back n generations whereas an undersegmented cell should not.
    from scipy.stats import mode
    import warnings
    warnings.filterwarnings("ignore")
    time, label = np.array(cell_id.split('_')).astype(int)
    younger_list = []
    current_cell = np.array([cell_id])
    while True:
        if n_generations == 0 or time == regionprops['$time'].max():
            break
        daughters = np.unique(np.hstack(regionprops.loc[current_cell]['daughter(s)'].values))
        current_cell = daughters
        n_generations -= 1
        time += 1
        if len(daughters) > 0:
            younger_list.append(daughters)
        else:
            break
    return younger_list


def forward_track_exhaustive(regionprops,
                             cell_id,
                             additional_search=5,
                             max_n_daughter=4):
    # the backtrack algorithm is based on the naive hypothesis that the fragments of an oversegmented cell
    # should have a common ancester tracing back n generations whereas an undersegmented cell should not.
    from scipy.stats import mode
    import warnings
    warnings.filterwarnings("ignore")
    time, label = np.array(cell_id.split('_')).astype(int)
    younger_list = []
    current_cell = np.array([cell_id])
    start_count = False
    while True:
        if time == regionprops['$time'].max():
            break
        if additional_search == 0:
            break
        daughters = np.unique(np.hstack(regionprops.loc[current_cell]['daughter(s)'].values))
        current_cell = daughters
        time += 1
        if len(daughters) >= max_n_daughter:
            start_count = True
        if start_count:
            additional_search -= 1
        if len(daughters) > 0:
            younger_list.append(daughters)
        else:
            break
    return younger_list


def _link_cells(tracker):
    from scipy.stats import mode
    import warnings
    warnings.filterwarnings("ignore")
    init_stat = {}
    mother_list = []
    daughter_list = []
    for t in range(len(tracker.frames) - 1):
        linking_matrix, trans_matrix, overlap_stat, labels1, labels2 = silly_link(tracker, t)
        for i, l in enumerate(labels1):
            k = '{}_{}'.format(t, l)
            if k not in init_stat:
                init_stat[k] = [[], []]
            daughters = ['{}_{}'.format(t + 1, labels2[j]) for j in np.where(linking_matrix[i] > 0)[0]]
            init_stat[k][1] = daughters
        for j, l in enumerate(labels2):
            k = '{}_{}'.format(t + 1, l)
            if k not in init_stat:
                init_stat[k] = [[], []]
            mothers = ['{}_{}'.format(t, labels1[i]) for i in np.where(linking_matrix[:, j] > 0)[0]]
            init_stat[k][0] = mothers
        tracker.frames[t].regionprops['$Tracker_ID'] = ['{}_{}'.format(t, l) for l in
                                                        tracker.frames[t].regionprops.index]
        tracker.frames[t].regionprops['$time'] = t
    tracker.frames[-1].regionprops['$Tracker_ID'] = ['{}_{}'.format(t + 1, l) for l in
                                                     tracker.frames[-1].regionprops.index]
    tracker.frames[-1].regionprops['$time'] = t + 1
    merged_rp = pd.concat([f.regionprops for f in tracker.frames]).set_index('$Tracker_ID').copy()
    mother_lists = [[init_stat[k][0], len(init_stat[k][0])] for k in merged_rp.index]
    daughter_lists = [[init_stat[k][1], len(init_stat[k][1])] for k in merged_rp.index]
    merged_rp[['mother(s)', 'n_mother']] = mother_lists
    merged_rp[['daughter(s)', 'n_daughter']] = daughter_lists
    tracker.regionprops = merged_rp
    return None


def area_ratio(areas):
    a1, a2 = areas
    if a1 > a2:
        ratio = a2 / a1
    else:
        ratio = a1 / a2
    return ratio


def show_highlight(cell_id, frames):
    t, label = np.array(cell_id.split('_')).astype(int)
    mask = (frames[t].labeled_mask > 0) * 1
    mask[frames[t].labeled_mask == label] = 3
    fig = plt.figure(figsize=(10, 10))
    plt.imshow(mask)


def overlap_match(frames, regionprops,
                  min_overlap_threshold=0.5,
                  min_overlap_area=100,
                  min_size_similarity=0.75,
                  min_cell_size=200):
    # link cells by overlap
    init_stat = {k: [[], []] for k in regionprops.index}
    mother_list = []
    daughter_list = []
    rp = regionprops.copy()
    overlap_stat = []
    for t in range(0, len(frames) - 1):
        df = _calculate_overlap_matrix(frames[t].labeled_mask,
                                       frames[t + 1].labeled_mask, t, t + 1)
        filtered_df = df[
            np.max(df[['frame1_overlap_frac', 'frame2_overlap_frac']].values, axis=1) > min_overlap_threshold]
        for (cell1, cell2) in filtered_df[['frame1_id', 'frame2_id']].values:
            init_stat[cell1][1] += [cell2]
            init_stat[cell2][0] += [cell1]
        overlap_stat.append(df)

    overlap_stat = pd.concat(overlap_stat)

    # find missing mother(s) for cells that moved a bit
    orphans_candidates = rp[(rp['$time'] > 0) & (rp['$touching_edge'] == 0)].index
    for o in orphans_candidates:
        if len(init_stat[o][0]) == 0 and len(init_stat[o][1]) > 0:
            subset = overlap_stat[overlap_stat['frame2_id'] == o].values
            missing_mother = []
            for s in subset:
                cond1 = s[2] > min_overlap_area
                cond2 = area_ratio((s[3], s[4])) > min_size_similarity
                cond3 = min(s[3], s[4]) > min_cell_size
                if cond1 * cond2 * cond3:
                    missing_mother += [s[0]]
                    init_stat[s[0]][1] += [o]
            init_stat[o][0] = missing_mother
    rp[['mother(s)', 'n_mother']] = [[np.unique(init_stat[k][0]), len(np.unique(init_stat[k][0]))] for k in rp.index]
    rp[['daughter(s)', 'n_daughter']] = [[np.unique(init_stat[k][1]), len(np.unique(init_stat[k][1]))] for k in
                                         rp.index]
    return rp, overlap_stat


def _calculate_overlap_matrix(frame1_labeled_mask,
                              frame2_labeled_mask,
                              frame1_label,
                              frame2_label):
    f1, f2 = frame1_labeled_mask.ravel(), frame2_labeled_mask.ravel()
    f1f2 = np.array([f1, f2]).T[f1 * f2 != 0]
    f1_counts = _unique2dict1D(f1)
    f2_counts = _unique2dict1D(f2)
    neighbors, overlap = np.unique(f1f2, return_counts=True, axis=0)
    f1_areas = np.array([f1_counts[i] for i in neighbors[:, 0]])
    f2_areas = np.array([f2_counts[i] for i in neighbors[:, 1]])

    f1_id = np.array(['{}_{}'.format(frame1_label, i) for i in neighbors[:, 0]])
    f2_id = np.array(['{}_{}'.format(frame2_label, i) for i in neighbors[:, 1]])
    overlap_matrix = np.hstack([f1_id.reshape(-1, 1),
                                f2_id.reshape(-1, 1),
                                overlap.reshape(-1, 1),
                                f1_areas.reshape(-1, 1),
                                f2_areas.reshape(-1, 1),
                                (overlap / f1_areas).reshape(-1, 1),
                                (overlap / f2_areas).reshape(-1, 1),
                                (overlap / (f1_areas + f2_areas - overlap)).reshape(-1, 1)])
    overlap_df = pd.DataFrame()
    overlap_df['frame1_id'] = f1_id
    overlap_df['frame2_id'] = f2_id
    overlap_df['overlap_area'] = overlap
    overlap_df['frame1_area'] = f1_areas
    overlap_df['frame2_area'] = f2_areas
    overlap_df['frame1_overlap_frac'] = overlap / f1_areas
    overlap_df['frame2_overlap_frac'] = overlap / f2_areas
    overlap_df['iou'] = overlap / (f1_areas + f2_areas - overlap)
    return overlap_df


def _unique2dict1D(array, nonzero=True):
    copied = array.copy()
    if nonzero:
        copied = copied[copied != 0]
    vals, counts = np.unique(copied, return_counts=True)
    return {v: c for v, c in zip(vals, counts)}


def _locate_spurious_events(regionprops,
                            n_generations=10,
                            window_size=5,
                            verbose=True):
    from scipy.stats import mode
    import warnings
    warnings.filterwarnings("ignore")
    to_be_merged = []
    to_be_split = []

    for cell in regionprops[regionprops['n_mother'] >= 2].index:
        back_track_record = np.flip(back_track(regionprops, cell, n_generations=n_generations), axis=0)
        forward_track_record = forward_track_exhaustive(regionprops, cell, max_n_daughter=2, additional_search=10)

        back_divergence = [len(x) for x in back_track_record]
        forward_divergence = [len(x) for x in forward_track_record]
        trace = list(back_track_record) + [np.array([cell])] + list(forward_track_record)
        counts = np.array(back_divergence + [1] + forward_divergence).astype(float)
        window_size = min(int(len(counts) / 2) * 2 + 1, window_size)
        if len(counts) >= window_size:
            convolved_counts = counts.copy()
            pad = int(window_size / 2)
            convolved_counts[pad:-pad] = np.convolve(counts, np.ones(window_size) / window_size, mode='valid')
            convolved_counts = np.round(convolved_counts).astype(int)
            # to_be_merged.append(convolved_counts)
            # to_be_split.append(counts)

            l = int(len(convolved_counts) / 2)
            v1 = convolved_counts[:l].mean()
            v2 = convolved_counts[-l:].mean()

            if v1 < 0.6 * v2:  # there should be a division event
                try:
                    turn_point = find_turn_point(convolved_counts)
                except:
                    turn_point = -1
                if turn_point >= 0:
                    if counts[:(turn_point + 1)].mean() < 1.5:
                        for i in np.where(counts[:(turn_point + 1)] > 1)[0]:
                            to_be_merged.append(trace[i].astype("<U22"))

                        for j in \
                        np.where((convolved_counts[(turn_point + 1):] <= 2) & (counts[(turn_point + 1):] == 1))[0]:
                            to_be_split.append(trace[j + turn_point + 1])
                    else:
                        for j in \
                        np.where((convolved_counts[:(turn_point + 1)] <= 2) & (counts[:(turn_point + 1)] == 1))[0]:
                            to_be_split.append(trace[j].astype("<U22"))
            else:  # not detected division event
                count_Mode = mode(counts)[0][0]
                if count_Mode == 1:
                    for i in np.where(counts > 1)[0]:
                        if len(trace[i]) == 2:
                            to_be_merged.append(trace[i].astype("<U22"))
                elif count_Mode == 2:
                    for j in np.where(counts == 1)[0]:
                        to_be_split.append(trace[j].astype("<U22"))
    new_split = []
    for x in to_be_split:
        x = x[0]
        if x not in new_split:
            new_split.append(x)
    new_merge = []
    for x in to_be_merged:
        x = list(x)
        if x not in new_merge:
            new_merge.append(x)
    # to_be_split = np.unique(np.array(to_be_split).astype("<U22"))
    # to_be_merged = np.unique(np.array(to_be_merged).astype("<U22"), axis=0)
    return new_split, new_merge


def find_turn_point(data, kernel=np.array([0.5, 0.5, 0, -0.5, -0.5])):
    pad = int(len(kernel) / 2)
    convolved = np.convolve(data, kernel, mode='valid')
    convolved = np.pad(convolved, ((pad, pad),))
    turn_point = np.where(convolved == 1)[0][0]
    return turn_point


def generate_correlation_map(x, y):
    """Correlate each n with each m.
    @abcd
    @https://stackoverflow.com/questions/30143417/computing-the-correlation-coefficient-between-two-multi-dimensional-arrays/30145770#30145770
    Parameters
    ----------
    x : np.array
      Shape N X T.

    y : np.array
      Shape M X T.

    Returns
    -------
    np.array
      N X M array in which each element is a correlation coefficient.

    """
    mu_x = x.mean(1)
    mu_y = y.mean(1)
    n = x.shape[1]
    if n != y.shape[1]:
        raise ValueError('x and y must ' +
                         'have the same number of timepoints.')
    s_x = x.std(1, ddof=n - 1)
    s_y = y.std(1, ddof=n - 1)
    cov = np.dot(x,
                 y.T) - n * np.dot(mu_x[:, np.newaxis],
                                   mu_y[np.newaxis, :])
    return cov / np.dot(s_x[:, np.newaxis], s_y[np.newaxis, :])


def transition_matrix(frames, t,
                      moments=['moments_hu-0', 'moments_hu-1', 'moments_hu-2',
                               'moments_hu-3', 'moments_hu-4', 'moments_hu-5', 'moments_hu-6'],
                      features=['area', 'major_axis_length', 'minor_axis_length', 'extent', 'aspect_ratio'],
                      normalize=True):
    """
    calcluate transition matrix
    channel0: Intersect(A[t],A[t+1])/Union(A[t],A[t+1])
    channel1: Intersect(A[t],A[t+1])/A[t]
    channel2: Intersect(A[t],A[t+1])/A[t+1]
    channel3: hu moments correlation matrix
    channel4: feature difference matrix
    channel5: centroid distance matrix
    channel6: nearest neighbor

    :params frames: momia.core.CellTracker.frames object (list)
    :params time: time
    :params features: rotation invariant features used to estimate pairwise correlation
    :return estimated transtion matrix and the two cell_id to index referenc dictionaries.
    """
    from sklearn.preprocessing import StandardScaler as ss
    f1 = frames[t]
    f2 = frames[t + 1]
    n1 = len(f1.regionprops)
    n2 = len(f2.regionprops)
    f1name_dict = {'{}_{}'.format(t, x): i for i, x in enumerate(f1.regionprops.index)}
    f2name_dict = {'{}_{}'.format(t + 1, x): i for i, x in enumerate(f2.regionprops.index)}
    area1 = f1.regionprops['area'].values
    area2 = f2.regionprops['area'].values

    # weights by overlap, channels 0-2
    transMatrix = np.zeros((n1, n2, 7))
    overlap_stat = _calculate_overlap_matrix(f1.labeled_mask, f2.labeled_mask, t, t + 1)
    for idx1, idx2, iou, frac1, frac2 in overlap_stat[
        ['frame1_id', 'frame2_id', 'iou', 'frame1_overlap_frac', 'frame2_overlap_frac']].values:
        transMatrix[f1name_dict[idx1], f2name_dict[idx2], 0] = iou
        transMatrix[f1name_dict[idx1], f2name_dict[idx2], 1] = frac1
        transMatrix[f1name_dict[idx1], f2name_dict[idx2], 2] = frac2

    # weights by hu moments
    m1 = f1.regionprops[moments].values
    m2 = f2.regionprops[moments].values

    if normalize:
        norm_v = ss().fit_transform(np.vstack([m1, m2]))
        m1 = norm_v[:n1]
        m2 = norm_v[-n2:]
    corr = generate_correlation_map(m1, m2)
    corr[corr < 0] = 0
    transMatrix[:, :, 3] = corr

    v1 = f1.regionprops[features].values
    v2 = f2.regionprops[features].values
    transMatrix[:, :, 4] = difference_matrix(v1, v2)

    cent1 = np.array([np.mean(x, axis=0) for x in f1.regionprops['$coords'].values])
    cent2 = np.array([np.mean(x, axis=0) for x in f2.regionprops['$coords'].values])
    dist = linalg.distance_matrix(cent1, cent2)
    argmin_dist = np.argmin(dist, axis=1)
    transMatrix[:, :, 5] = dist_prob(dist)
    transMatrix[:, :, 6] = 1 - (
                2 * np.abs(area1[:, np.newaxis] - area2[np.newaxis]) / (area1[:, np.newaxis] + area2[np.newaxis]))
    return transMatrix, f1name_dict, f2name_dict, overlap_stat


def silly_link(tracker_obj, t,
               iou_threshold=0.7,
               overlap_threshold=0.7,
               similarity_threshold=0.9,
               diff_threshold=0.2,
               logprob_threshold=4,
               min_iou_threshold=0.1):
    frames = tracker_obj.frames
    trans_matrix, f1n, f2n, overlap_stat = transition_matrix(frames, t, normalize=True)
    f1 = frames[t]
    f2 = frames[t + 1]
    n1 = len(f1.regionprops)
    n2 = len(f2.regionprops)
    labels1 = f1.regionprops.index
    labels2 = f2.regionprops.index
    matched_l1 = np.zeros(n1)
    matched_l2 = np.zeros(n2)
    area1 = f1.regionprops['area'].values
    area2 = f2.regionprops['area'].values

    linking_matrix = np.zeros((n1, n2))
    cond1 = trans_matrix[:, :, 0] > iou_threshold
    cond2 = trans_matrix[:, :, 4] > similarity_threshold
    cond3 = trans_matrix[:, :, 6] > similarity_threshold
    linking_matrix[cond1 & cond2 & cond3] = 1
    matched_l1[np.sum(linking_matrix, axis=1) > 0] = 1
    matched_l2[np.sum(linking_matrix, axis=0) > 0] = 1
    f1_remnant = np.where(matched_l1 == 0)[0]
    f2_remnant = np.where(matched_l2 == 0)[0]

    log_prob_matrix = np.sum(np.log2(trans_matrix + 1), axis=-1)
    log_dist_matrix = np.sum(np.log2(trans_matrix[:, :, np.array([0, 1, 2, 5])] + 1), axis=-1)

    # fig=plt.figure(figsize=(4,4))
    # plt.imshow(linking_matrix)
    for i in f1_remnant:
        cond1 = log_prob_matrix[i] > logprob_threshold
        cond2 = matched_l2 == 0
        cond3 = trans_matrix[i, :, 0] > min_iou_threshold
        cond4 = trans_matrix[i, :, 6] > similarity_threshold
        future_self_id = np.where(cond1 & cond2 & cond3 & cond4)[0]
        if len(future_self_id) >= 1:
            j = future_self_id[np.argmax(log_prob_matrix[i, future_self_id])]
            linking_matrix[i, j] = 1
            matched_l1[i] = 1
            matched_l2[j] = 1

    f1_remnant = np.where(matched_l1 == 0)[0]
    # cases when 2 or more particles on frame t map to one particle on frame t+1
    for i in f1_remnant:
        cond1 = trans_matrix[i, :, 1] > overlap_threshold
        cond2 = trans_matrix[i, :, 1] > trans_matrix[i, :, 2]
        cond3 = matched_l2 != 1
        future_self_id = np.where(cond1 & cond2 & cond3)[0]
        if len(future_self_id) == 1:
            j = future_self_id[0]
            matched_l1[i] = 1
            matched_l2[j] = 2  # N->1 event, note that the sibling particle won't necessarily overlap with j
            linking_matrix[i, j] = 1

    # if areas match, linkage is complete
    for j in np.where(matched_l2 == 2)[0]:
        f2_area = area2[j].sum()
        f1_ids = np.where(linking_matrix[:, j] == 1)[0]
        if len(f1_ids) > 1:
            f1_area = area1[f1_ids].sum()
            if np.abs(2 * (f1_area - f2_area) / (f1_area + f2_area)) < diff_threshold:
                matched_l2[j] = 1
        elif len(f1_ids) == 1:
            l1_1 = f1_ids[0]
            l1_1_coords = f1.regionprops.loc[labels1[l1_1], '$coords']
            for l1_2 in np.flip(np.argsort(log_dist_matrix[:, j])):
                cond1 = l1_2 != l1_1
                cond2 = matched_l1[l1_2] == 0
                # l2_1 and l2_2 should be sitting next to each other
                if cond1 and cond2:
                    l1_2_coords = f1.regionprops.loc[labels1[l1_2], '$coords']
                    if linalg.distance_matrix(l1_1_coords, l1_2_coords).min() < 2:
                        f1_area = area1[np.array([l1_1, l1_2])].sum()
                        if np.abs(2 * (f1_area - f2_area) / (f1_area + f2_area)) < diff_threshold:
                            matched_l1[l1_2] = 1
                            matched_l2[j] = 1
                            linking_matrix[l1_2, j] = 1
                            break

    # cases when 2 or more particles on frame t+1 map to one particle on frame t
    f2_remnant = np.where(matched_l2 == 0)[0]
    for j in f2_remnant:
        cond1 = trans_matrix[:, j, 2] > overlap_threshold
        cond2 = trans_matrix[:, j, 1] < trans_matrix[:, j, 2]
        cond2 = matched_l1 != 1
        prev_self_id = np.where(cond1 & cond2)[0]
        if len(prev_self_id) == 1:
            i = prev_self_id[0]
            matched_l1[i] = 2  # 1->N event, note that the sibling particle won't necessarily overlap with i
            matched_l2[j] = 1
            linking_matrix[i, j] = 1

    for i in np.where(matched_l1 == 2)[0]:
        f2_ids = np.where(linking_matrix[i, :] == 1)[0]
        f1_area = area1[i].sum()
        # if areas match, linkage is complete
        if len(f2_ids) > 1:
            f2_area = area2[f2_ids].sum()
            if np.abs(2 * (f1_area - f2_area) / (f1_area + f2_area)) < diff_threshold:
                matched_l1[i] = 1

        # find sibling, then if areas match, linkage is complete
        elif len(f2_ids) == 1:
            l2_1 = f2_ids[0]
            l2_1_coords = f2.regionprops.loc[labels2[l2_1], '$coords']
            for l2_2 in np.flip(np.argsort(log_dist_matrix[i])):
                cond1 = l2_2 != l2_1
                cond2 = matched_l2[l2_2] == 0
                # l2_1 and l2_2 should be sitting next to each other
                if cond1 and cond2:
                    l2_2_coords = f2.regionprops.loc[labels2[l2_2], '$coords']
                    if linalg.distance_matrix(l2_1_coords, l2_2_coords).min() < 2:
                        f2_area = area2[np.array([l2_1, l2_2])].sum()
                        if np.abs(2 * (f1_area - f2_area) / (f1_area + f2_area)) < diff_threshold:
                            matched_l1[i] = 1
                            matched_l2[l2_2] = 1
                            linking_matrix[i, l2_2] = 1
                            break

    f1_remnant = np.where(matched_l1 == 0)[0]
    f2_remnant = np.where(matched_l2 == 0)[0]
    # vsnap or any movement can be really annoying
    for i in f1_remnant:
        for j in np.flip(np.argsort(log_dist_matrix[i]))[:3]:
            if j in f2_remnant:
                if trans_matrix[i, j, 6] > similarity_threshold - 0.1:
                    matched_l1[i] = 1
                    matched_l2[j] = 1
                    linking_matrix[i, j] = 1
                    break
                elif area1[i] > area2[j]:
                    matched_l2[j] = 1
                    linking_matrix[i, j] = 1
                    f2_area = area2[np.where(linking_matrix[i] > 0)].sum()
                    if np.abs(2 * (area1[i] - f2_area) / (area1[i] + f2_area)) < diff_threshold:
                        matched_l1[i] = 1
                        break
                elif area1[i] < area2[j]:
                    matched_l1[i] = 1
                    linking_matrix[i, j] = 1
                    f1_area = area1[np.where(linking_matrix[:, j] > 0)].sum()
                    if np.abs(2 * (area2[j] - f1_area) / (area2[j] + f1_area)) < diff_threshold:
                        matched_l2[j] = 1
                        break

    # now we deal with the remaining unassigned particles
    f1_remnant = np.where(matched_l1 == 0)[0]
    f2_remnant = np.where(matched_l2 == 0)[0]

    for i in f1_remnant:
        if len(f2_remnant) > 0:
            feature_maxj = f2_remnant[np.argmax(trans_matrix[i, f2_remnant, 4])]
            moments_maxj = f2_remnant[np.argmax(trans_matrix[i, f2_remnant, 3])]
            if feature_maxj == moments_maxj and feature_maxj in f2_remnant:
                if trans_matrix[i, feature_maxj, 4] > similarity_threshold:
                    feature_maxi = f1_remnant[np.argmax(trans_matrix[f1_remnant, feature_maxj, 4])]
                    moments_maxi = f1_remnant[np.argmax(trans_matrix[f1_remnant, feature_maxj, 3])]
                    if feature_maxi == moments_maxi == i:
                        linking_matrix[i, j] = 1

    return linking_matrix, trans_matrix, overlap_stat, labels1, labels2


import numpy as np
import pandas as pd
import tifffile
import read_roi as rr
from skimage import filters, measure
from skimage.morphology import disk, remove_small_objects, binary_opening, binary_dilation, binary_erosion
from ..utils import linalg,generic

__all__ = ['roi2multilabel','roi2multimasks',
           'prediction2seed','prediction2foreground','np_method',
           'dist2labels_simp','dimension_scaler']

def roi2multilabel(roi_file, image_shape, dst, erosion_radius=2, subdivide_degree=2):
    from skimage import draw
    from skimage.measure import subdivide_polygon
    roi = rr.read_roi_zip(roi_file)
    canvas = np.zeros(image_shape)
    core = np.zeros(image_shape)
    for k, v in roi.items():
        xy = np.array([v['y'], v['x']]).T
        xy = subdivide_polygon(xy, degree=subdivide_degree, preserve_ends=True)
        mask = draw.polygon2mask(image_shape, xy)
        eroded = binary_erosion(mask, disk(erosion_radius))
        core[eroded == 1] = 1
    dilated = binary_dilation(core, disk(erosion_radius + 1))
    canvas[(dilated - core) > 0] = 125
    canvas[core > 0] = 254
    tifffile.imwrite(dst, canvas.astype(np.uint8), imagej=True)


def roi2multimasks(roi_file, image_file, dst, erosion_radius=2, subdivide_degree=2):
    from skimage import draw, measure
    from skimage.segmentation import watershed
    from skimage.measure import subdivide_polygon
    roi = rr.read_roi_zip(roi_file)
    image = tifffile.imread(image_file)
    image_shape = image.shape
    canvas = np.zeros(image_shape)
    core = np.zeros(image_shape)
    for k, v in roi.items():
        xy = np.array([v['y'], v['x']]).T
        xy = subdivide_polygon(xy, degree=subdivide_degree, preserve_ends=True)
        mask = draw.polygon2mask(image_shape, xy)
        eroded = binary_erosion(mask, disk(erosion_radius))
        core[eroded == 1] = 1
        canvas[mask == 1] = 1
    labeled = watershed(image, markers=measure.label(core),
                        mask=canvas, watershed_line=False, compactness=100)
    tifffile.imwrite(dst, labeled.astype(np.uint16), imagej=True)


def prediction2seed(pred_multilabel_mask,
                    seed_min=0.3,
                    edge_max=0.75,
                    min_seed_size=20,
                    opening_radius=1):
    seed = (pred_multilabel_mask[:, :, 2] > seed_min) * (pred_multilabel_mask[:, :, 1] < edge_max) * 1
    if isinstance(opening_radius, int) and opening_radius > 0:
        seed = binary_opening(seed, disk(opening_radius))
    seed = remove_small_objects(seed, min_seed_size)
    return measure.label(seed)


def dist2labels_simp(dist, mask,
                     dist_threshold=0.25,
                     mask_threshold=0.8,
                     opening=2,
                     min_particle_size=10,
                     watershedline=False):
    from skimage import measure, segmentation
    basin = 1 - dist
    seed = measure.label(dist > dist_threshold)
    seed = remove_small_objects(seed, min_size=min_particle_size)
    binary_mask = mask > mask_threshold
    if int(opening) > 0:
        binary_mask = binary_opening(binary_mask, disk(int(opening)))
    watersheded = segmentation.watershed(basin,
                                         markers=seed,
                                         mask=binary_mask,
                                         compactness=100,
                                         watershed_line=watershedline)
    return watersheded


def prediction2foreground(pred_multilabel_mask, channel=0,
                          threshold=0.4,
                          erosion_radius=1):
    fg = (pred_multilabel_mask[:, :, channel] < threshold) * 1
    # fg=morphology.binary_closing(fg,morphology.disk(1))
    if isinstance(erosion_radius, int) and erosion_radius > 0:
        fg = binary_erosion(fg, disk(erosion_radius))
    return fg * 1


def compute_dist(multi_mask, smooth_factor=10, max_v=7):
    from skimage import measure
    import edt
    from scipy.ndimage import median_filter

    canvas = np.zeros(multi_mask.shape)

    for i in np.unique(multi_mask):
        if i > 0:
            u_mask = (multi_mask == i) * 1
            xy = np.array(np.where(u_mask > 0)).T
            c = measure.find_contours(median_filter(multi_mask == i, 3), level=0.5)
            if len(c) == 1:
                smoothed = linalg.spline_approximation(c[0], smooth_factor=smooth_factor, n=len(c[0]))
                dis = np.min(linalg.distance_matrix(xy, smoothed), axis=1)
            else:
                dis = edt.edt(u_mask)[xy[:, 0], xy[:, 1]]
            canvas[xy[:, 0], xy[:, 1]] = dis
    return generic.normalize_image(canvas, min_perc=0, max_perc=100, min_v=0, max_v=7)


def np_method(data, method='mean', **kwargs):
    if method in ['default', 'mean', 'average', 'MEAN', 'Mean', 'Average']:
        return np.mean(data, **kwargs)
    elif method in ['median', 'Median', 'MEDIAN']:
        return np.median(data, **kwargs)
    elif method in ['Max', 'max', 'MAX']:
        return np.max(data, **kwargs)
    elif method in ['Min', 'min', 'MIN']:
        return np.min(data, **kwargs)
    else:
        print('method not found, use np.mean instead')
        return np.mean(data, **kwargs)


def dimension_scaler(image, axis=(0, 3)):
    """
    image used for segment is reformated as a four-dimensional tensor with each dimension representing:
    #0 - number of classes
    #1,2 - 2D image
    #3 - number of channels (e.g. this should be 3 for RGB images)

    if class labels are provided for a single channel image, the axis paramter should be (3).
    if a multi-channel image is provided with no class labesl, the axis paramter should be (0).
    """

    if len(image.shape) == 4:
        return image
    elif len(image.shape) == 3 and len(axis) == 1:
        return np.expand_dims(image, axis=axis)
    elif len(image.shape) == 2 and len(axis) == 2:
        return np.expand_dims(image, axis=axis)
    else:
        raise ValueError("Image shape and provided axis information don't match")


import numpy as np
from scipy.interpolate import splprep, splev, RectBivariateSpline
from numba import jit
import warnings


def spline_approximation(n2array, n=200, smooth_factor=1, closed=True):
    warnings.filterwarnings("ignore")
    """
    Updated: 2023.04.12
    Compute a spline approximation of a 2D curve.

    :param n2array: a 2D numpy array of shape (n_points, 2) representing the curve to be approximated
    :param n: number of points of the output polyline/polygon
    :param smooth_factor: smoothing factor for the spline
    :param closed: whether the curve is closed or not
    :return: a 2D numpy array of shape (n, 2) representing the spline approximation of the curve
    """

    if closed:
        tck, u = splprep(n2array.T, u=None, s=smooth_factor, per=1)
    else:
        tck, u = splprep(n2array.T, u=None, s=smooth_factor)
    u_new = np.linspace(u.min(), u.max(), n)
    x_new, y_new = splev(u_new, tck, der=0)
    return np.array([x_new, y_new]).T


def bilinear_interpolate_numpy(im, x, y):
    """
    bilinear interpolation 2D
    :param im: target image
    :param x: x coordinates
    :param y: y coordinates
    :return: interpolated data
    """
    h,l = im.shape
    padded = np.zeros((h+1,l+1))
    padded[:h,:l] += im
    im = padded
    x0 = x.astype(int)
    x1 = x0 + 1
    y0 = y.astype(int)
    y1 = y0 + 1
    Ia = im[x0,y0]
    Ib = im[x0,y1]
    Ic = im[x1,y0]
    Id = im[x1,y1]
    wa = (x1-x) * (y1-y)
    wb = (x1-x) * (y-y0)
    wc = (x-x0) * (y1-y)
    wd = (x-x0) * (y-y0)
    return (Ia*wa) + (Ib*wb) + (Ic*wc) + (Id*wd)


def unit_perpendicular_vector(data, closed=True):
    """
    Compute the unit perpendicular vector of a set of 2D points.

    :param data: A numpy array of shape (n, 2) representing the 2D points.
    :param closed: A boolean indicating whether the curve is closed or not.
    :return: A numpy array of shape (n, 2) representing the unit perpendicular vector of the curve.
    """
    p1 = data[1:]
    p2 = data[:-1]
    dxy = p1 - p2
    ang = np.arctan2(dxy.T[1], dxy.T[0]) + 0.5 * np.pi
    dx, dy = np.cos(ang), np.sin(ang)
    unit_dxy = np.array([dx, dy]).T
    if not closed:
        unit_dxy = np.concatenate([[unit_dxy[0]], unit_dxy])
    else:
        unit_dxy = np.concatenate([unit_dxy, [unit_dxy[0]]])
    return unit_dxy


def line_contour_intersect_matrix(line, contour, orthogonal_vectors=None):

    """
    Compute the intersection point(s) between a line and a contour.
    Previously named "intersect_matrix"

    :param line: A numpy array of shape (2, 2) representing the start and end points of a straight line.
    :param contour: A numpy array of shape (n, 2) representing the points of the contour.
    :param orthogonal_vectors: A numpy array of shape (n, 2) representing the unit perpendicular vectors of the contour.
    :return: A tuple of numpy arrays representing the x and y coordinates of the intersection point(s).
    """

    if orthogonal_vectors is None:
        dxy = unit_perpendicular_vector(line, closed=False)
    else:
        dxy = orthogonal_vectors
    v1, v2 = contour[:-1], contour[1:]
    x1, y1 = v1.T
    x2, y2 = v2.T
    x3, y3 = line.T
    perp_xy = line + dxy
    x4, y4 = perp_xy.T
    A1 = y2 - y1
    B1 = x1 - x2
    C1 = A1 * x1 + B1 * y1
    A2 = y4 - y3
    B2 = x3 - x4
    C2 = A2 * x3 + B2 * y3

    A1B2 = A1[:, np.newaxis] * B2
    A1C2 = A1[:, np.newaxis] * C2
    B1A2 = B1[:, np.newaxis] * A2
    B1C2 = B1[:, np.newaxis] * C2
    C1A2 = C1[:, np.newaxis] * A2
    C1B2 = C1[:, np.newaxis] * B2

    intersect_x = (B1C2 - C1B2) / (B1A2 - A1B2)
    intersect_y = (A1C2 - C1A2) / (A1B2 - B1A2)
    return intersect_x, intersect_y


@jit(nopython=True, cache=True)
def intersect_lines(x1, y1, x2, y2, x3, y3, x4, y4):
    """
    Compute the intersection point between two lines.

    :param x1: A float representing the x-coordinate of the first point of the first line.
    :param y1: A float representing the y-coordinate of the first point of the first line.
    :param x2: A float representing the x-coordinate of the second point of the first line.
    :param y2: A float representing the y-coordinate of the second point of the first line.
    :param x3: A float representing the x-coordinate of the first point of the second line.
    :param y3: A float representing the y-coordinate of the first point of the second line.
    :param x4: A float representing the x-coordinate of the second point of the second line.
    :param y4: A float representing the y-coordinate of the second point of the second line.
    :return: A tuple of floats representing the x and y coordinates of the intersection point.
    """
    A1 = y2 - y1
    B1 = x1 - x2
    C1 = A1 * x1 + B1 * y1
    A2 = y4 - y3
    B2 = x3 - x4
    C2 = A2 * x3 + B2 * y3
    intersect_x = (B1 * C2 - B2 * C1) / (A2 * B1 - A1 * B2)
    intersect_y = (A1 * C2 - A2 * C1) / (B2 * A1 - B1 * A2)
    return intersect_x, intersect_y


def line_contour_intersection(p1, p2, contour):
    """
    Compute the intersection point(s) between a line and a contour.

    :param p1: A tuple of floats representing the (x, y) coordinates of the start point of the line.
    :param p2: A tuple of floats representing the (x, y) coordinates of the end point of the line.
    :param contour: A numpy array of shape (n, 2) representing the points of the contour.
    :return: A numpy array of shape (m, 2) representing the x and y coordinates of the intersection point(s).
    """
    v1, v2 = contour[:-1], contour[1:]
    x1, y1 = v1.T
    x2, y2 = v2.T
    x3, y3 = p1
    x4, y4 = p2
    xy = np.array(intersect_lines(x1, y1, x2, y2, x3, y3, x4, y4)).T
    dxy_v1 = xy - v1
    dxy_v2 = xy - v2
    dxy = dxy_v1 * dxy_v2
    intersection_points = xy[np.where(np.logical_and(dxy[:, 0] < 0, dxy[:, 1] < 0))]
    if len(intersection_points) > 2:
        dist = np.sum(np.square(np.tile(p1, (len(intersection_points), 1)) - intersection_points),
                      axis=1)
        intersection_points = intersection_points[np.argsort(dist)[0:2]]
    return intersection_points

def distance(v1, v2):
    """
    Euclidean distance of two points
    :param v1: vecorized coordinate of the first point.
    :param v2: vecorized coordinate of the second point.
    :return: distance between two points.
    """
    return np.sqrt(np.sum((np.array(v1) - np.array(v2)) ** 2))


def distance_matrix(data1, data2):
    """
    Compute the Euclidean distance matrix between two sets of 2D points.

    :param data1: A numpy array of shape (n, 2) representing the first set of 2D points.
    :param data2: A numpy array of shape (m, 2) representing the second set of 2D points.
    :return: A numpy array of shape (n, m) representing the Euclidean distance matrix between the two sets of points.
    """
    x1, y1 = data1.T
    x2, y2 = data2.T
    dx = x1[:, np.newaxis] - x2
    dy = y1[:, np.newaxis] - y2
    dxy = np.sqrt(dx ** 2 + dy ** 2)
    return dxy


def point_line_orientation(l1,l2,p):
    """
    Compute the orientation of a point with respect to a line.

    :param l1: A tuple of floats representing the (x, y) coordinates of the first point of the line.
    :param l2: A tuple of floats representing the (x, y) coordinates of the second point of the line.
    :param p: A tuple of floats representing the (x, y) coordinates of the point.
    :return: A float representing the orientation of the point with respect to the line.
    """
    return np.sign(((l2[0] - l1[0]) * (p[1] - l1[1]) - (l2[1] - l1[1]) * (p[0] - l1[0])))


def interpolate_2Dmesh(data_array, smooth=1):
    """
    Interpolate a 2D mesh of data using a rectangular bivariate spline.

    :param data_array: A numpy array of shape (n, m) representing the 2D mesh of data to interpolate.
    :param smooth: A float representing the smoothing factor to use for the spline interpolation.
    :return: A RectBivariateSpline object representing the interpolated mesh of data.
    """
    x_mesh = np.linspace(0, data_array.shape[0] - 1, data_array.shape[0]).astype(int)
    y_mesh = np.linspace(0, data_array.shape[1] - 1, data_array.shape[1]).astype(int)
    return RectBivariateSpline(x_mesh, y_mesh, data_array, s=smooth)


def bend_angle_closed(data, window=3, deg=True):
    """
    Compute the bending angle of a closed 2D curve.

    :param data: A numpy array of shape (n, 2) representing the 2D curve.
    :param window: An integer representing the size of the window to use for computing the bending angle.
    :param deg: A boolean indicating whether to return the angle in degrees (True) or radians (False).
    :return: A float representing the bending angle of the curve.
    """
    p1 = np.concatenate((data[-window:],data[:-window])).T
    p2 = data.copy().T
    p3 = np.concatenate((data[window:],data[0:window])).T
    p1p2 = p1[0]*1+p1[1]*1j - (p2[0]*1+p2[1]*1j)
    p1p3 = p1[0]*1+p1[1]*1j - (p3[0]*1+p3[1]*1j)
    return np.angle(p1p3/p1p2, deg=deg)


def bend_angle_open(data, window=3, deg=True):
    """
    Compute the bending angle of an open 2D curve.

    :param data: A numpy array of shape (n, 2) representing the 2D curve.
    :param window: An integer representing the size of the window to use for computing the bending angle.
    :param deg: A boolean indicating whether to return the angle in degrees (True) or radians (False).
    :return: A float representing the bending angle of the curve.
    """
    p1 = data[:-2*window].T
    p2 = data[window:-window].T
    p3 = data[2*window:].T
    p1p2 = p1[0]*1+p1[1]*1j - (p2[0]*1+p2[1]*1j)
    p2p3 = p2[0]*1+p2[1]*1j - (p3[0]*1+p3[1]*1j)
    return np.angle(p2p3/p1p2, deg=deg)


def angle_between_vectors(v1, v2):
    """
    Compute the angle between two vectors.

    :param v1: A numpy array representing the first vector.
    :param v2: A numpy array representing the second vector.
    :return: A float representing the angle between the two vectors in radians.
    """
    d1 = np.sqrt(np.sum(v1 ** 2))
    d2 = np.sqrt(np.sum(v2 ** 2))
    return np.arccos(np.dot(v1, v2) / (d1 * d2))


def measure_length(data, pixel_microns=1):
    """
    Compute the length of a 2D polyline.

    :param data: A numpy array of shape (n, 2) representing the 2D polyline.
    :param pixel_microns: A float representing the conversion factor from pixels to microns.
    :return: A float representing the length of the polyline in microns.
    """
    v1,v2 = data[:-1], data[1:]
    length = np.sqrt(np.sum((np.array(v1) - np.array(v2)) ** 2, axis=1)).sum()*pixel_microns
    return(length)


def line_contour_intersect_matrix(line, contour, orthogonal_vectors=None):

    """
    Compute the intersection point(s) between a line and a contour.
    Previously named "intersect_matrix"

    :param line: A numpy array of shape (2, 2) representing the start and end points of a straight line.
    :param contour: A numpy array of shape (n, 2) representing the points of the contour.
    :param orthogonal_vectors: A numpy array of shape (n, 2) representing the unit perpendicular vectors of the contour.
    :return: A tuple of numpy arrays representing the x and y coordinates of the intersection point(s).
    """

    if orthogonal_vectors is None:
        dxy = unit_perpendicular_vector(line, closed=False)
    else:
        dxy = orthogonal_vectors
    v1, v2 = contour[:-1], contour[1:]
    x1, y1 = v1.T
    x2, y2 = v2.T
    x3, y3 = line.T
    perp_xy = line + dxy
    x4, y4 = perp_xy.T
    A1 = y2 - y1
    B1 = x1 - x2
    C1 = A1 * x1 + B1 * y1
    A2 = y4 - y3
    B2 = x3 - x4
    C2 = A2 * x3 + B2 * y3

    A1B2 = A1[:, np.newaxis] * B2
    A1C2 = A1[:, np.newaxis] * C2
    B1A2 = B1[:, np.newaxis] * A2
    B1C2 = B1[:, np.newaxis] * C2
    C1A2 = C1[:, np.newaxis] * A2
    C1B2 = C1[:, np.newaxis] * B2

    intersect_x = (B1C2 - C1B2) / (B1A2 - A1B2)
    intersect_y = (A1C2 - C1A2) / (A1B2 - B1A2)
    return intersect_x, intersect_y





__author__ = 'jz-rolling'

import os, glob, pickle as pk
import numpy as np
import tifffile
import nd2reader as nd2

__all__=['ImageLoader','sort2folder','save_model','load_model','imread']

class ImageLoader:

    def __init__(self, image_file=None, log=None, verbose=False):
        self.header = ''
        self.format = ''
        self.source = ''
        self.pixel_microns = 1
        self.channels = []
        self.n_channels = 1
        self.n_fields = 1
        self.n_frames = 1
        self.axes = 'PTCZYX'
        self.axes_dict = {}
        self._update_axes_dict()
        self.shape = (2048, 2048)
        self.stack_shape = ()
        self.volumes = {}
        self.external_obj = None
        self.metadata = {}
        self.log = log
        self.verbose = verbose
        if image_file is not None:
            self.read(image_file)

    def read(self, image_file):
        """
        Load image file.
        :param image_file: file path.
        :return: self.
        """
        if not os.path.isfile(image_file):
            raise ValueError('File {} not found.'.format(image_file))
        if image_file.endswith('.nd2'):
            self._softread_nd2file(image_file)
        elif image_file.endswith('.tif') or image_file.endswith('.tiff'):
            self._softread_tiffile(image_file)
        else:
            raise ValueError('.{} format not supported (yet)!'.format(image_file.split('.')[-1]))

    def get_slice(self,
                  channel=0,
                  position=0,
                  time=0,
                  zplane=0):
        """
        get image slice based on user provided specifications
        :param channel: channel index, between 0 and self.n_channels-1.
        :param position: position index, between 0 and self.n_positions-1.
        :param time: time frame index, between 0 and self.n_frames-1.
        :param zplane: Z-plane index, between 0 and self.n_zpositions-1.
        :return: single image slice which has a shape of self.shape
        """

        if self.format == 'tif':
            sliced_img = _get_slice_tifffile(self, self.axes, channel=channel, time=time, position=position,
                                             zplane=zplane)
        elif self.format == 'nd2':
            sliced_img = np.array(self.external_obj.get_frame_2D(c=channel, v=position, t=time, z=zplane))
        else:
            raise ValueError('Data format {} not supported (yet)!'.format(self.format))
        return sliced_img

    def get_slice_all_channels(self, position=0,time=0,zplane=0):
        """
        get a multi-channel image slice.
        :param position: position index, between 0 and self.n_positions-1.
        :param time: time frame index, between 0 and self.n_frames-1.
        :param zplane: Z-plane index, between 0 and self.n_zpositions-1.
        :return: a dictionary of images with the keys being the items in self.channels and
                 each value being a channnel image which has a shape of self.shape
        """
        channel_dict = {}
        for i,c in enumerate(self.channels):
            channel_dict[c] = self.get_slice(position=position,time=time,zplane=zplane,channel=i)
        return channel_dict

    def _softread_tiffile(self, image_file, header=None, channels=None, pixel_microns=0.065):
        """
        open .tif file/hyperstack, current version can only handle single clips or hyperstack rendered by imagej.
        :param image_file: .tif file path
        :param header: in case you want to use a different file header
        :param channels: microscopy channel names
        :param pixel_microns: micron length per pixel
        """

        # convert .tif file to standard OMEGA input
        # only supports single view, single/multi- channel tif data
        if not isinstance(image_file, str):
            raise ValueError('Illegal input!')
        if not image_file.endswith('.tif'):
            raise ValueError('Input {} is not a .tif file.'.format(image_file.split('/')[-1]))
        if header is None:
            header = image_file.split('/')[-1].split('.')[0]

        #
        img = tifffile.TiffFile(image_file)
        series = img.series[0]

        self.format = 'tif'
        self.header = header
        self.axes = series.axes
        self._update_axes_dict()
        self.shape = (series.shape[self.axes_dict['Y']], series.shape[self.axes_dict['X']])

        if 'Info' in img.imagej_metadata:
            self.metadata = imagej_meta2dict(img.imagej_metadata['Info'])
        else:
            self.metadata = img.imagej_metadata

        # use embedded scaling factor if available
        self.pixel_microns = float(self.metadata['dCalibration']) if 'dCalibration' in self.metadata else float(
            pixel_microns)
        self.n_channels = series.shape[self.axes_dict['C']] if 'C' in self.axes else 1

        # imageJ hyperstack does not support images with multiple fields x multiple z positions
        self.n_fields = series.shape[self.axes_dict['Z']] if 'Z' in self.axes else 1
        self.n_zpositions = series.shape[self.axes_dict['Z']] if 'Z' in self.axes else 1

        # now support time frames
        self.n_frames = series.shape[self.axes_dict['T']] if 'T' in self.axes else 1
        self.stack_shape = series.shape

        if channels is not None:
            if len(channels) != self.n_channels:
                raise ValueError('{} channels found, {} channel names specified!'.format(self.n_channels, len(channels)))
            self.channels = channels
        else:
            self.channels = ['C{}'.format(i + 1) for i in range(self.n_channels)]

        self.external_obj = img

    def _softread_nd2file(self, image_file, header=None):
        """
        open .nd2 image file
        :param image_file: .nd2 file path
        :param header: in case you want to use a different file header
        """
        # convert .nd2 file to standard OMEGA input
        # img axes: v->t->c->y->x
        if not isinstance(image_file, str):
            raise ValueError('Illegal input!')
        if not image_file.endswith('.nd2'):
            raise ValueError('Input {} is not a .nd2 file.'.format(image_file.split('/')[-1]))
        if header is None:
            header = image_file.split('/')[-1].split('.')[0]
        # sorted_data = {}

        img = nd2.ND2Reader(image_file)
        sizes = img.sizes
        metadata = img.metadata
        shape = (sizes['x'], sizes['y'])
        channels = metadata['channels']
        n_channels = len(channels)
        n_fields = sizes['v'] if 'v' in sizes else 1
        n_time = sizes['t'] if 't' in sizes else 1
        n_zcoords = sizes['z'] if 'z' in sizes else 1
        self.format = 'nd2'
        self.header = header
        self.metadata = metadata
        self.pixel_microns = metadata['pixel_microns']
        self.external_obj = img
        self.channels = channels
        self.n_fields = n_fields
        self.n_channels = n_channels
        self.n_timepoints = n_time
        self.n_zpositions = n_zcoords
        self.shape = shape
        self.stack_shape = (self.n_fields, self.n_timepoints, self.n_channels,
                            self.n_zpositions, self.shape[0], self.shape[1])

    def _update_axes_dict(self):
        self.axes_dict = {x: i for i, x in enumerate(self.axes)}

def sort2folder(src_folder):
    """
    sort JOBS .nd2 files to well folders
    :param src_folder:
    :return: move .nd2 files to well folders.
    """
    files = sorted(glob.glob(src_folder + '*.nd2'))
    for f in files:
        header = f.split('/')[-1]
        well = header.split('_')[1][4:]
        subfolder = '{}{}/'.format(src_folder, well)
        if not os.path.isdir(subfolder):
            os.mkdir(subfolder)
        os.rename(f, '{}{}'.format(subfolder, header))


def imagej_meta2dict(ImageJinfo):
    """
    Extract metadata from imageJ processsed .tif files
    :param ImageJinfo: imagej converted metadata
    :return: dictionary of metadata
    """
    if "\n" in ImageJinfo:
        ImageJinfo = ImageJinfo.split("\n")
    ImageJdict = {}
    for info in ImageJinfo:
        info = info.split(" = ")
        if len(info) == 2:
            key = info[0]
            val = info[1]
            if key.startswith(" "):
                key = key[1:]
            if val.startswith(" "):
                val = val[1:]
            ImageJdict[key] = val
    return ImageJdict


def _get_slice_tifffile(img_stack, axes, position=None, time=None, zplane=None, channel=None):
    # Create a dictionary to map axes to their indices
    axes_dict = {axis: i for i, axis in enumerate(axes[:-2])}  # ignore the XY axes

    # Initialize an empty slice
    img_slice = [slice(None)] * (len(axes) - 2)  # ignore the XY axes

    # Set the slices according to the input parameters
    if position is not None and 'P' in axes_dict:
        img_slice[axes_dict['P']] = position
    if time is not None and 'T' in axes_dict:
        img_slice[axes_dict['T']] = time
    if channel is not None and 'C' in axes_dict:
        img_slice[axes_dict['C']] = channel
    if zplane is not None and 'Z' in axes_dict:  # z-plane is always considered even when not requested
        img_slice[axes_dict['Z']] = zplane

    # Calculate the number of elements before the desired slice
    flat_index = 0
    for i, slice_ in enumerate(img_slice):
        if slice_ != slice(None):
            flat_index += slice_ * np.prod(img_stack.stack_shape[i + 1:-2])  # ignore the XY axes
    # Return the index in the flattened image stack
    return img_stack.external_obj.pages[int(flat_index)].asarray()


def save_model(model, pickle_filename):
    pk.dump(model, open(pickle_filename, 'wb'))


def load_model(pickle_filename):
    return pk.load(open(pickle_filename, 'rb'))

def imread(file):
    return tifffile.imread(file)


from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler, LabelEncoder
from sklearn.ensemble import RandomForestClassifier
import pandas as pd
import pickle as pk
from ..utils import *


class GenericClassifier:

    def __init__(self,
                 method='RandomForest',
                 existing_classifier=None,
                 **kwargs):

        self.model = None
        if method == 'RandomForest':
            self.model = RandomForestClassifier(**kwargs)
        if method == 'PretrainedModel' and existing_classifier is not None:
            self.model = existing_classifier(**kwargs)
        self.features = []
        self.method = method
        self.normalizer = StandardScaler()
        self.z_transform = False

    def train(self, training_df, test_size=0.4, random_state=33, verbose=True, normalize=False, **kwargs):

        # check feature presence
        _missing_features = []
        _features = []
        for f in self.features:
            if f not in training_df.columns:
                _missing_features.append(f)
            else:
                _features.append(f)

        if len(_missing_features) > 0 and verbose:
            print('The following features are not found in the training dataset: {}'.format(','.join(_missing_features)))
            self.features = _features

        # must have $annotation
        if '$annotation' not in training_df.columns:
            raise ValueError("Dataframe contains no '$annotation' for classification.")

        self.data = training_df[self.features].values
        self.data[np.where(np.isnan(self.data))] = 0
        if normalize:
            self.z_transform = True
            self.normalizer.fit(self.data)
            self.data = self.normalizer.transform(self.data)
        self.annotations = training_df['annotation'].values.reshape(-1, 1)

        # train test split
        self.X_train, \
        self.X_test, \
        self.Y_train, \
        self.Y_test = train_test_split(self.data, self.annotations, test_size=test_size, random_state=random_state)

        self.model.fit(self.X_train, self.Y_train.ravel())
        self.classification_score = self.model.score(self.X_test, self.Y_test.ravel())

        if verbose:
            print('Classification score using {} model is {}'.format(self.method, self.classification_score))

    def predict(self, target_df, probability=False):
        _missing_features = []
        for f in self.features:
            if f not in target_df.columns:
                _missing_features.append(f)
        if len(_missing_features) > 0:
            raise ValueError(
                'The following features are not found in the training dataset: {}'.format(','.join(_missing_features)))
        if len(target_df) == 0:
            raise ValueError('No data found!')
        data = target_df[self.features].values
        data = np.nan_to_num(data, copy=True, nan=0.0, posinf=0, neginf=0)
        if len(data) == 1:
            data = data.reshape(1, -1)
        if self.z_transform:
            data = self.normalizer.transform(data)
        if probability:
            return self.model.predict_proba(data)
        else:
            return self.model.predict(data)

    def update(self):
        return None








# Import necessary modules
from setuptools import setup, find_packages

# Define package name, version, and the author who doesn't know a thing about coding
name = 'MOMIA2'
version = '0.0.2.dev3'
author = 'jz-rolling'
author_emails = ['juzhu@hsph.harvard.edu', 'zhujh@im.ac.cn']

# Define package description
description = 'Mycobacteria-optimized microscopy image analysis version 2 (sort of)'

# Load README content
with open("README.md", "r", encoding = "utf-8") as fh:
    long_description = fh.read()

# Define package requirements
requirements = [
    'numpy==1.25.2',
    'trackpy==0.6.1',
    'pandas==2.1.1',
    'scikit-image==0.21.0',
    'scikit-learn==1.3.1',
    'scipy==1.11.3',
    'matplotlib>=3.7',
    'networkx==3.1',
    'numba==0.58.0',
    'tifffile==2023.9.26',
    'nd2reader==3.3.0',
    'networkx==3.1',
    'seaborn==0.13.0',
    'opencv-python==4.8.1.78',
    'Pillow==10.0.1',
    'PyWavelets==1.4.1',
    'PyYAML==6.0.1',
    'read-roi==1.6.0',
    'tqdm==4.66.1',
]

# Define package classifiers
classifiers = [
    'License :: OSI Approved :: MIT License',
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    'Operating System :: OS Independent'
]

# Define package setup
setup(
    name=name,
    version=version,
    author=author,
    author_email=', '.join(author_emails),
    description=description,
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    classifiers=classifiers,
    packages = find_packages(
        where='MOMIA2'
    ),
    package_dir={"": "MOMIA2"},
    python_requires = ">=3.8"
)


from skimage import feature, measure, draw
from matplotlib import pyplot as plt, cm
from matplotlib.gridspec import GridSpec as gs
from ..utils import *

__all__ = ['find_puncta','puncta_heatmaps']

def find_puncta(img,mask,min_distance=1, threshold_rel=0.1):
    """

    :param img:
    :param mask:
    :param min_distance:
    :param threshold_rel:
    :return:
    """
    img_smoothed = filters.gaussian(img, sigma=0.5)
    img_smoothed = (img_smoothed - img_smoothed.min()) / (img_smoothed.max() - img_smoothed.min())
    LoG = ndi.gaussian_laplace(img_smoothed, sigma=1)
    extrema = feature.peak_local_max(-LoG * mask, min_distance=min_distance, threshold_rel=threshold_rel)
    output = []

    for ex in extrema:
        x, y = ex
        newx, newy, stable_maxima = subpixel_approximation_quadratic(x, y, img_smoothed)
        if stable_maxima:
            # z_image = local_puncta_zscore(img_smoothed, ex)
            # z_LoG = local_puncta_zscore(LoG, ex)
            output.append([newx, newy, img[x, y], LoG[x, y]])
    return np.array(output)


def subpixel_approximation_quadratic(x, y, energy_map, max_iteration=2):
    """

    :param x:
    :param y:
    :param energy_map:
    :param max_iteration:
    :return:
    """

    local_maxima_found = False
    for i in range(max_iteration):
        cube = energy_map[x - 1:x + 2, y - 1:y + 2]
        if cube[1, 1] == cube.max():
            local_maxima_found = True
            vx0, vx1, vx2 = cube[0, 1], cube[1, 1], cube[2, 1]
            vy0, vy1, vy2 = cube[1, 0], cube[1, 1], cube[1, 2]
            x += quadratic_maxima_approximation(vx0, vx1, vx2)
            y += quadratic_maxima_approximation(vy0, vy1, vy2)
            break
        else:
            dx, dy = np.array(np.unravel_index(cube.argmax(), cube.shape)) - 1
            x += dx
            y += dy
    return x, y, local_maxima_found


def quadratic_maxima_approximation(q1, q2, q3):
    """

    :param q1:
    :param q2:
    :param q3:
    :return:
    """
    if q2 <= max(q1, q3):
        raise ValueError('Mid point must be local maxima!')
    else:
        return (0.5 * (q1 - q3)) / (q1 - 2 * q2 + q3)


def local_puncta_zscore(image, coords, trim_low_bound=5):
    """

    :param image:
    :param coords:
    :param trim_low_bound:
    :return:
    """

    x, y = coords
    data = np.sort((image[x - 4:x + 5, y - 4:y + 5].copy()).flatten())[trim_low_bound:]

    threshold = filters.threshold_isodata(data, nbins=64)
    threshold = 0.5 * (threshold + data[-4:].mean())

    fg = data[data > threshold]
    bg = data[data <= threshold]

    if len(fg) <= 5:
        z = 0
    else:
        z = (np.mean(fg) - np.mean(bg)) / bg.std()
    return z


def standardize_punclist(punc):
    if len(punc) == 0:
        return np.array([None] * 6)
    else:
        return punc


def create_canvas(width=101, height=600):
    """

    :param width:
    :param height:
    :return:
    """
    canvas = np.zeros((width, height))
    r = int(width / 2)
    rr1, cc1 = draw.ellipse(r, int(r * 1.6), r, int((r) * 1.6))
    rr2, cc2 = draw.ellipse(r, int(height - 1.6 * r), r, int((r) * 1.6))
    rr3, cc3 = draw.rectangle(start=(1, int(r * 1.6)), end=(width - 2, height - int(r * 1.6)),
                              shape=canvas.shape)
    canvas[rr3, cc3] = 1
    canvas[rr1, cc1] = 1
    canvas[rr2, cc2] = 1
    l = len(np.nonzero(np.sum(canvas, axis=0))[0])
    counter = 0
    canvas = canvas.T
    return canvas


def create_mesh(canvas):
    """

    :param canvas:
    :return:
    """
    xt, yt = np.nonzero(canvas)
    l, m = np.sum(canvas, axis=0), np.sum(canvas, axis=1)
    norm_yt = np.zeros(xt.shape)
    norm_xt = (xt - xt.min()) / (l.max() - 1)
    count = 0
    for i in range(len(xt)):
        r = m.max() / m[xt[i]]
        norm_yt[i] = count * r / (m.max() - 1)
        if i != len(xt) - 1:
            if xt[i + 1] == xt[i]:
                count += 1
            else:
                count = 0
        else:
            count += 1
    return (xt, yt, norm_xt, norm_yt)


def project_image(xt, yt, norm_xt, norm_yt, canvas, data):
    """

    :param xt:
    :param yt:
    :param norm_xt:
    :param norm_yt:
    :param canvas:
    :param data:
    :return:
    """
    paint = np.zeros(canvas.shape)
    xid = norm_xt.copy()
    yid = norm_yt.copy()
    xid *= data.shape[0] - 1
    yid *= data.shape[1] - 1
    interpolated = bilinear_interpolate_numpy(data, xid, yid)
    paint[xt, yt] = interpolated
    return (paint)


def initiate_projection():
    """

    :return:
    """
    width = 75
    heights = [250, 350, 450, 550, 650]
    pad = np.zeros((50, width))
    gap = np.zeros((750, 10))
    padded = [gap]
    xt_list, yt_list, nxt_list, nyt_list = [], [], [], []
    for i in range(len(heights)):
        a = create_canvas(width=width, height=heights[i])
        half_pad = np.tile(pad, (len(heights) - i, 1))
        m_pad = np.concatenate([half_pad, a, half_pad], axis=0)
        m_pad = np.concatenate([m_pad, gap], axis=1)
        xt, yt, norm_xt, norm_yt = create_mesh(m_pad)
        xt_list.append(xt)
        yt_list.append(yt)
        nxt_list.append(norm_xt)
        nyt_list.append(norm_yt)
        padded.append(m_pad)
    contours = measure.find_contours(np.concatenate(padded, axis=1), level=0)
    optimized_outline = []
    for contour in contours:
        optimized_outline.append(spline_approximation(contour, n=2 * len(contour)))
    return padded, xt_list, yt_list, nxt_list, nyt_list, optimized_outline


def standardize_punclist(punc):
    """

    :param punc:
    :return:
    """
    if len(punc) == 0:
        return np.array([None] * 6)
    else:
        return punc


def filter_pundata(punc, threshold):
    if None in punc:
        return punc
    else:
        return standardize_punclist(punc[punc[:, 2] >= threshold])


def punc_string_parser(punc_data):
    """

    :param punc_data:
    :return:
    """
    punc = str(punc_data)
    new_list = []
    counter = 0
    substring = ''
    for x in str(punc):
        if x not in '1234567890e-+.':
            if counter != 0:
                new_list.append(substring)
                substring = ''
                counter = 0
        else:
            substring += x
            counter += 1
    new_list = np.array(new_list).reshape(int(len(new_list) / 6), 6).astype(float)
    return new_list


def adjust_relative_width(x, y, normx, normy):
    """

    :param x:
    :param y:
    :param normx:
    :param normy:
    :return:
    """
    p1 = np.sum((normx[np.newaxis, :] - y[:, np.newaxis]) < 0, axis=1)
    p0 = p1 - 1
    w = normy[p0] + (normy[p1] - normy[p0]) * (y - normx[p0]) / (normx[p1] - normx[p0])
    return x * w


def puncta_heatmaps(df, optimized_outline, padded, coords, coord2w_list,
                    channels=['TRITC', 'CY5'], targets=['1', '2'], cutoffs=[300, 100]):
    """

    :param df:
    :param optimized_outline:
    :param padded:
    :param coords:
    :param coord2w_list:
    :param channels:
    :param targets:
    :param cutoffs:
    :return:
    """
    from scipy.stats import gaussian_kde
    mRNA_loc = plt.figure(figsize=(4 * len(channels) + 0.2, 11.2))
    grids = gs(2, len(channels))
    for k, c in enumerate(channels):
        target = targets[k]
        cutoff = cutoffs[k]
        ax1 = mRNA_loc.add_subplot(grids[0, k])
        ax2 = mRNA_loc.add_subplot(grids[1, k])
        fg = np.concatenate(padded, axis=1)

        bg_scatter, bg_heatmap = np.zeros(fg.shape), np.zeros(fg.shape)
        ax1.imshow(bg_scatter, cmap='Greys', aspect='auto')
        ax1.set_xlim(0, bg_scatter.shape[1] - 1)
        for outline in optimized_outline:
            ax1.plot(outline.T[1], outline.T[0], c="black", alpha=1, lw=0.5)
            ax2.plot(outline.T[1], outline.T[0], c="white", alpha=1, lw=0.5)
        median_lengths = []
        for i, frac in enumerate([0, 0.2, 0.4, 0.6, 0.8]):
            l1 = df['length'].quantile(frac)
            l2 = df['length'].quantile(frac + 0.2)
            subset = df[(df['length'] < l2) & (df['length'] >= l1)]
            median_lengths.append(subset['length'].median())
            punc_list = []
            for punc in subset['{}_puncta_data'.format(c)]:
                punc = str(punc)
                if 'None' not in punc:
                    punc_list.append(punc_string_parser(punc))
            if len(punc_list) > 2:
                punc_list = np.vstack(punc_list)
                if len(punc_list) > 2:
                    punc_list = punc_list[punc_list[:, 2] >= cutoff]
            if len(punc_list) > 2:
                v, x, y = punc_list[:, np.array([2, 4, 5])].T
                norm_v = (v) / np.percentile(v, 85)
                norm_v[norm_v < 0] = 0
                norm_v[norm_v > 1] = 1
                normx, normy = coord2w_list[i]
                x0, y0, l, w = coords[4 - i]

                x = adjust_relative_width(x, y, normx, normy)
                # plot scatter
                ax1.scatter(x * w + x0, (y - 0.5) * l + y0, s=norm_v * 20, alpha=0.5, fc=cm.get_cmap('Blues')(norm_v))

                # plot heatmap
                data = np.array([x * w + x0, (y - 0.5) * l + y0]).T
                xgrid, ygrid, X, Y = construct_grids(x0, w, y0, l, 1)
                kernel = gaussian_kde(data.T, weights=v)
                positions = np.vstack([X.ravel(), Y.ravel()])
                Z = np.reshape(kernel(positions).T, X.shape)

                # normalize density
                Z /= Z.max()
                bg_heatmap[Y, X] = Z
        bg_heatmap[fg == 0] = 0
        ax2.imshow(bg_heatmap, cmap='magma', aspect='auto')
        ax1.set_xticks(np.array(coords)[:, 0])
        ax1.set_xticklabels(np.flip(np.round(median_lengths, 1)), fontsize=14)
        ax1.set_xlabel('median length [μm]', fontsize=14)
        ax2.set_xticks(np.array(coords)[:, 0])
        ax2.set_xticklabels(np.flip(np.round(median_lengths, 1)), fontsize=14)
        ax2.set_xlabel('median length [μm]', fontsize=14)
        ax1.set_title('{}-$\it{}$'.format(c, target), fontsize=14)
        ax1.set_yticks([])
        ax2.set_yticks([])
    return mRNA_loc


def construct_grids(x0, w, y0, l, gridsize=1):
    xmin = int(x0 - w * 0.5)
    xmax = int(x0 + w * 0.5)
    ymin = int(y0 - l * 0.5)
    ymax = int(y0 + l * 0.5)
    xgrid = np.arange(xmin, xmax, gridsize)
    ygrid = np.arange(ymin, ymax, gridsize)
    X, Y = np.meshgrid(xgrid, ygrid)
    return xgrid, ygrid, X, Y


import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.gridspec import GridSpec as gs

__all__ = ['demograph','grouped_scatter','grouped_box_plot']

def demograph(sorted_midline,
              sorted_length,
              n=256,
              max_l=8,
              align='center',
              reorient=True,
              forced_reorientation=False,
              reorientation_sequence=[],
              uniformed=False,
              skip_end=0,
              smooth_window=0):
    demograph = []
    _flip_sequence = []

    if forced_reorientation and len(reorientation_sequence) != len(sorted_midline):
        raise ValueError("Lengths don't match!")

    for i, ml in enumerate(sorted_midline):
        _flipped = False
        l = sorted_length[i]
        frac = min(1, l / max_l)
        half_n = int(0.5 * n * frac)
        pad = int(n * 0.5) - half_n
        if smooth_window > 0 and smooth_window % 2 == 1:
            ml = np.convolve(ml, np.ones(smooth_window) / smooth_window, mode='same')
        if skip_end > 0:
            ml = ml[skip_end:-skip_end]

        if reorient:
            if not forced_reorientation:
                ml, _flipped = reorient_data(ml)
            else:
                _flipped = bool(reorientation_sequence[i])
                if _flipped:
                    ml = np.flip(ml)
        _flip_sequence.append(_flipped)

        if not uniformed:
            interpolated = np.interp(x=np.linspace(0, 1, 2 * half_n), xp=np.linspace(0, 1, len(ml)), fp=ml)
            if align == 'center':
                arr = np.pad(interpolated, (pad, pad), 'constant')
            elif align == 'left':
                arr = np.pad(interpolated, (0, 2 * pad), 'constant')
            elif align == 'right':
                arr = np.pad(interpolated, (2 * pad, 0), 'constant')
        else:
            arr = np.interp(x=np.linspace(0, 1, n), xp=np.linspace(0, 1, len(ml)), fp=ml)
        demograph.append(arr)
    return np.array(demograph), np.array(_flip_sequence)


def reorient_data(data):
    half_l = int(round(len(data) / 2))
    flip = bool(data[:half_l].mean() < data.mean())
    if flip:
        data = np.flip(data)
    return data, flip


def grouped_scatter(merged_df, feature1, feature2):
    colors = {'Rv1830-mut': cm.get_cmap('Set2')(1), 'WT': cm.get_cmap('Set2')(0)}
    vmax1 = merged_df[feature1].quantile(0.995) * 1.2
    vmin1 = merged_df[feature1].quantile(0.01) * 0.4
    vmax2 = merged_df[feature2].quantile(0.995) * 1.2
    vmin2 = merged_df[feature2].quantile(0.01) * 0.4
    if feature1 == 'FITC_mean':
        vmin1 -= 100
    if feature2 == 'FITC_mean':
        vmin2 -= 100
    fig = plt.figure(figsize=(16, 12))
    grids = gs(8, 8, hspace=0.1, wspace=0.1)
    for i, r in enumerate('CDEF'):
        rowd = merged_df[merged_df['Row'] == r]
        for j, w in enumerate(sorted(list(rowd['Well'].unique()))):
            if r == 'C':
                j += int(j / 2) * 2
            ax = fig.add_subplot(grids[j, i])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlim(vmin1, vmax1)
            ax.set_ylim(vmin2, vmax2)
            welld = rowd[rowd['Well'] == w]
            time = welld['Time'].values[0]
            treatment = welld['INH Conc.'].values[0]
            strain = welld['Strain'].values[0]
            if j == 0:
                ax.set_title('{} hours'.format(time))
            if j % 2 == 0:
                ax.text(0.05, 0.9, '{}MIC'.format(treatment), transform=ax.transAxes, va='center', fontweight='bold')
            if i == 0 and j == 0:
                ax.set_ylabel(feature2, fontsize=14)
            if i == 1 and j == 7:
                ax.set_xlabel(feature1, fontsize=14)
            sns.scatterplot(x=welld[feature1],
                            y=welld[feature2],
                            s=10, alpha=0.3, fc=colors[strain], ax=ax)
            sns.kdeplot(x=welld[feature1],
                        y=welld[feature2], levels=5, color="black", linewidths=1, ax=ax)

    legend_ax = fig.add_subplot(grids[2, 0])
    legend_ax.scatter(0.1, 0.8, s=50, alpha=0.8, fc=colors['Rv1830-mut'])
    legend_ax.scatter(0.1, 0.2, s=50, alpha=0.8, fc=colors['WT'])
    legend_ax.set_xlim(0, 1);
    legend_ax.set_ylim(0, 1);
    legend_ax.axis('off');
    legend_ax.text(0.18, 0.8, 'Rv1830-mut', fontsize=14, va='center', ha='left')
    legend_ax.text(0.18, 0.2, 'WT', fontsize=14, va='center', ha='left')


def grouped_hist(merged_df, feature1):
    colors = {'Rv1830-mut': cm.get_cmap('Set2')(1), 'WT': cm.get_cmap('Set2')(0)}
    vmax1 = merged_df[feature1].quantile(0.995) * 1.2
    vmin1 = merged_df[feature1].quantile(0.01) * 0.4

    fig = plt.figure(figsize=(16, 12))
    grids = gs(8, 8, hspace=0.1, wspace=0.1)
    for i, r in enumerate('CDEF'):
        rowd = merged_df[merged_df['Row'] == r]
        for j, w in enumerate(sorted(list(rowd['Well'].unique()))):
            if r == 'C':
                j += int(j / 2) * 2
            ax = fig.add_subplot(grids[j, i])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_xlim(vmin1, vmax1)
            welld = rowd[rowd['Well'] == w]
            time = welld['Time'].values[0]
            treatment = welld['INH Conc.'].values[0]
            strain = welld['Strain'].values[0]
            if j == 0:
                ax.set_title('{} hours'.format(time))
            if j % 2 == 0:
                ax.text(0.05, 0.9, '{}MIC'.format(treatment), transform=ax.transAxes, va='center', fontweight='bold')
            if i == 1 and j == 7:
                ax.set_xlabel(feature1, fontsize=14)
            ax.hist(x=welld[feature1], fc=colors[strain], bins=50, ec='black', lw=0.5, alpha=0.5)

    legend_ax = fig.add_subplot(grids[2, 0])
    legend_ax.scatter(0.1, 0.8, s=50, alpha=0.8, fc=colors['Rv1830-mut'])
    legend_ax.scatter(0.1, 0.2, s=50, alpha=0.8, fc=colors['WT'])
    legend_ax.set_xlim(0, 1);
    legend_ax.set_ylim(0, 1);
    legend_ax.axis('off');
    legend_ax.text(0.18, 0.8, 'Rv1830-mut', fontsize=14, va='center', ha='left')
    legend_ax.text(0.18, 0.2, 'WT', fontsize=14, va='center', ha='left')


def grouped_box_plot(merged_df, feature):
    vmax = merged_df[feature].quantile(0.995) * 1.5
    vmin = merged_df[feature].quantile(0.001) * 0.2
    fig = plt.figure(figsize=(10, 4))
    grids = gs(1, 7)
    for i, t in enumerate(np.sort(merged_df['Time'].unique())):
        if t == 0:
            ax = fig.add_subplot(grids[0, 0])
        else:
            ax = fig.add_subplot(grids[i * 2 - 1:i * 2 + 1])
        subset = merged_df[merged_df['Time'] == t]
        g = sns.boxplot(data=subset, x='INH Conc.', y=feature, hue='Strain', ax=ax, palette='Set2',
                        hue_order=['WT', 'Rv1830-mut'])
        if i != 3:
            ax.legend([], [])
        if i != 0:
            ax.set_yticks([])
            ax.set_ylabel('')
        else:
            ax.set_ylabel(feature, fontsize=12)
        if i != 2:
            ax.set_xlabel('')
        else:
            ax.set_xlabel('INH MIC', loc='left', fontsize=12)
        ax.set_title('{} hours'.format(t))
        ax.set_ylim(vmin, vmax);


from ..utils import *
from matplotlib import pyplot as plt

__all__ = ['plot_cells']

def plot_cells(patch,
            cell_ids=[],
            figsize=(5, 5),
            outline_prop={},
            midline_prop={},
            cell_plot=False,
            channel=-1):

    """
    Plots cells from a given patch with outlined boundaries and midlines (if available).

    :param patch: The patch containing cell data
    :type patch: Patch object
    :param cell_ids: List of cell ids to be plotted
    :type cell_ids: list, optional
    :param figsize: Size of the figure to be plotted
    :type figsize: tuple, optional
    :param outline_prop: Properties for the cell outline (color, lw - linewidth, ls - linestyle)
    :type outline_prop: dict, optional
    :param midline_prop: Properties for the cell midline (color, lw - linewidth, ls - linestyle)
    :type midline_prop: dict, optional
    :param cell_plot: If True, plot individual cell images, else plot all cells on patch
    :type cell_plot: bool, optional
    :param channel: Channel index for multi-channel images. If -1, use reference image
    :type channel: int, optional
    """

    outline_color = outline_prop['color'] if 'color' in outline_prop else 'orange'
    outline_lw = float(outline_prop['lw']) if 'lw' in outline_prop else 1
    outline_ls = outline_prop['ls'] if 'ls' in outline_prop else '-'
    midline_color = midline_prop['color'] if 'color' in midline_prop else 'orange'
    midline_lw = midline_prop['lw'] if 'lw' in midline_prop else 1
    midline_ls = midline_prop['ls'] if 'ls' in midline_prop else '-'

    if channel == -1:
        img = patch.get_ref_image()
    else:
        img = patch.get_channel_data(channel)

    if len(cell_ids) == 0:
        cell_ids = patch.regionprops.index[patch.regionprops['$include'] == 1]
    if cell_plot:
        for i in cell_ids:
            fig = plt.figure(figsize=figsize)
            if i in patch.regionprops.index:
                x1, y1, x2, y2, outline, midlines, raw_outline = patch.regionprops.loc[
                    i, ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2',
                        '$refined_outline', '$midlines', '$outline']].values
                plt.imshow(img[x1:x2, y1:y2], cmap='gist_gray')
                if len(outline) > 2:
                    plt.plot(outline[:, 1], outline[:, 0], color=outline_color, lw=outline_lw, ls=outline_ls)
                elif len(raw_outline) > 2:
                    plt.plot(raw_outline[:, 1], raw_outline[:, 0], color=outline_color, lw=outline_lw,
                             ls=outline_ls)
                if len(midlines) > 0:
                    for midline in midlines:
                        plt.plot(midline[:, 1], midline[:, 0],
                                 color=midline_color, lw=midline_lw, ls=midline_ls)
    else:
        fig = plt.figure(figsize=figsize)
        plt.imshow(img, cmap='gist_gray')
        for i in cell_ids:
            if i in patch.regionprops.index:
                x1, y1, x2, y2, outline, midlines, raw_outline = patch.regionprops.loc[
                    i, ['$opt-x1', '$opt-y1', '$opt-x2', '$opt-y2',
                        '$refined_outline', '$midlines', '$outline']].values
                if len(outline) > 2:
                    plt.plot(outline[:, 1] + y1, outline[:, 0] + x1, color=outline_color, lw=outline_lw,
                             ls=outline_ls)
                elif len(raw_outline) > 2:
                    plt.plot(raw_outline[:, 1] + y1, raw_outline[:, 0] + x1, color=outline_color,
                             lw=outline_lw, ls=outline_ls)
                if len(midlines) > 0:
                    for midline in midlines:
                        plt.plot(midline[:, 1] + y1, midline[:, 0] + x1, color=midline_color, lw=midline_lw,
                                 ls=midline_ls)


from .generic import *
from skimage import filters, transform
from scipy import fftpack
import warnings
import numpy as np


"""
=================================== Section 1. xy drift correction ===================================

"""
def drift_detection(img, reference, upsample_factor=10):
    """
    wrapped around phase_cross_correlation in skimage's registration module
    """
    from skimage.registration import phase_cross_correlation
    shift, error, _diff = phase_cross_correlation(reference,
                                                  img,
                                                  upsample_factor=upsample_factor)
    return shift

def drift_correction(target_img, drift, max_drift=1000):
    """
    Corrects xy drift between phase contrast image and fluorescent image(s).

    Parameters
    ----------
    target_img : ndarray
        Input image
    drift : tuple
        Subpixel xy drift
    max_drift : int, optional
        Maximum allowed drift, by default 1000

    Returns
    -------
    ndarray
        Drift corrected image
    """
    import scipy.ndimage as ndi

    # get target image dtype
    img_dtype = target_img.dtype

    # Fourier shift the image by the given subpixel xy drift
    offset_image = ndi.fourier_shift(np.fft.fftn(target_img), drift)
    offset_image = np.fft.ifftn(offset_image)
    offset_image = np.round(offset_image.real)
    offset_image[offset_image <= 0] = 1

    # Rescale the image to avoid overflow
    img_dtype = target_img.dtype
    if max(np.abs(drift)) <= max_drift:
        if img_dtype == np.uint8:
            offset_image[offset_image >= 255] = 255
            return offset_image.astype(np.uint8)
        elif img_dtype == np.uint16:
            offset_image[offset_image >= 65530] = 65530
            return offset_image.astype(np.uint16)
        elif img_dtype == int:
            offset_image[offset_image >= 2147483647] = 2147483647
        elif img_dtype == np.float:
            return offset_image.astype(np.float)
        elif img_dtype == np.uint12:
            offset_image[offset_image >= 4095] = 4095
            return offset_image.astype(np.uint12)
        else:
            print("Unsupported dtype: {}".format(img_dtype))
            return target_img
    else:
        return target_img

def get_xydrift(ref_img, target_img):
    from skimage import registration
    """
    Calculate the shift between two images using phase cross-correlation.

    Parameters
    ----------
    ref_img : ndarray
        Reference image.
    target_img : ndarray
        Target image.

    Returns
    -------
    shift : tuple
        Tuple of x and y shift values.

    Examples
    --------
    >>> shift = get_xydrift(ref_img, target_img)
    """
    shift, error, _diff = registration.phase_cross_correlation(ref_img, target_img, upsample_factor=10)
    return shift


def correct_xydrift_timeseries(timeseries,
                               ref_channel=-1,
                               order='XYTC',
                               verbose=False,
                               drift_correction_threshold=5):
    """
    Corrects xy drift of microscopy images in a time series.

    Parameters
    ----------
    timeseries : ndarray
        Input time series of images
    ref_channel : int, optional
        Reference channel for drift correction, by default 1
    order : str, optional
        Dimension order of the input time series, by default 'XYTC'
    verbose : bool, optional
        Whether to print progress messages, by default False
    drift_correction_threshold : int, optional
        Maximum allowed drift, by default 5

    Returns
    -------
    ndarray
        Drift corrected time series of images
    """

    from tqdm import tqdm
    if np.sum([(order.find(x) >= 0) * 1 for x in 'XYT']) == 3:
        if 'C' in order and len(order) == 4:
            new_order = 'TCXY'
        elif 'C' not in order and len(order) == 3:
            new_order = 'TXY'
        else:
            raise ValueError('image dimension order specified to be "{}", but {} dimensional image found.'.format(order,
                                                                                                                  len(timeseries.shape)))
    else:
        raise ValueError(
            'Illegal order found: {}. Dimension order should only consist to following letters: "X","Y","C","T".'.format(
                                                                                                                    order))

    order_dict = {x: i for i, x in enumerate(order)}
    tm = np.transpose(timeseries, tuple([order_dict[x] for x in new_order]))
    if len(new_order) == 3:
        tm = np.expand_dims(tm, axis=1)

    last_drift = np.array([0, 0])
    nframes = tm.shape[0]
    nchannels = tm.shape[1]
    max_driftx = 0
    max_drifty = 0

    drift_corrected = np.zeros(tm.shape)
    drift_corrected[0, :, :, :] = tm[0, :, :, :]

    if ref_channel >= nframes:
        raise ValueError('reference frame id (ref_frame) should be smaller than the number of channels of the image')

    if verbose:
        print('Correct for planar drift')
    for t in tqdm(range(1, nframes)):
        last_drift = last_drift + get_xydrift(tm[t - 1, ref_channel],
                                              tm[t, ref_channel])
        max_driftx = max(abs(last_drift[0]), max_driftx)
        max_drifty = max(abs(last_drift[1]), max_drifty)
        for c in range(nchannels):
            drift_corrected[t, c] = drift_correction(tm[t, c], last_drift,
                                                     drift_correction_threshold)
    max_driftx = int(max_driftx)
    max_drifty = int(max_drifty)

    cropped = drift_corrected[:, :, max_driftx:(drift_corrected.shape[2] - max_driftx),
              max_drifty:(drift_corrected.shape[3] - max_drifty)]
    if verbose:
        print('Drift correction complete')
        print('Output image shape is {}'.format(cropped.shape))
    return cropped


"""
=================================== Section 2. background correction ===================================

"""

class rolling_ball:
    """
    A class to create a rolling ball structuring element for background subtraction.

    :param radius: int, optional (default=50)
        The radius of the rolling ball.

    :ivar width: int
        The width of the rolling ball.
    :ivar shrink_factor: int
        The factor by which the ball is shrunk.
    :ivar arc_trim_per: int
        The percentage of the arc to be trimmed.
    :ivar radius: float
        The radius of the rolling ball divided by the shrink factor.
    :ivar ball: numpy.ndarray
        The rolling ball structuring element.
    """

    def __init__(self, radius=50):
        self.width = 0
        if radius <= 10:
            self.shrink_factor = 1
            self.arc_trim_per = 24
        elif radius <= 20:
            self.shrink_factor = 2
            self.arc_trim_per = 24
        elif radius <= 100:
            self.shrink_factor = 4
            self.arc_trim_per = 32
        else:
            self.shrink_factor = 8
            self.arc_trim_per = 40
        self.radius = radius / self.shrink_factor
        self.build()

    def build(self):
        """
        Builds the rolling ball structuring element.
        """
        x_trim = int(self.arc_trim_per * self.radius / 100)
        half_width = int(self.radius - x_trim)
        self.width = int(2 * half_width + 1)
        r_squre = np.ones((self.width, self.width)) * (self.radius ** 2)
        squared = np.square(np.linspace(0, self.width - 1, self.width) - half_width)
        x_val = np.tile(squared, (self.width, 1))
        y_val = x_val.T
        self.ball = np.sqrt(r_squre - x_val - y_val)


def rolling_ball_bg_subtraction(data, **kwargs):
    """
        Subtracts background from an image using rolling ball algorithm.

        Parameters
        ----------
        data : numpy.ndarray
            The input image.
        rolling_ball_radius : int, optional
            The radius of the rolling ball. Default is 40.
        **kwargs : dict
            Additional keyword arguments to be passed to the function.

        Returns
        -------
        numpy.ndarray
            The background-subtracted image.
    """
    rolling_ball_radius = kwarg_or_default('rolling_ball_radius', kwargs, 40)
    output_data = data.copy()
    smoothed = filters.gaussian(data, sigma=1) * 65535
    ball = rolling_ball(radius=rolling_ball_radius)
    shrunk_img = shrink_img_local_min(smoothed, shrink_factor=ball.shrink_factor)
    bg = rolling_ball_bg(shrunk_img, ball.ball)
    bg_rescaled = transform.resize(bg, data.shape, anti_aliasing=True)
    output_data = output_data - bg_rescaled
    output_data[output_data <= 0] = 0
    return output_data.astype(np.uint16)


@jit(nopython=True, cache=True)
def shrink_img_local_min(image, shrink_factor=4):
    s = shrink_factor
    r, c = image.shape[0], image.shape[1]
    r_s, c_s = int(r / s), int(c / s)
    shrunk_img = np.ones((r_s, c_s))
    for x in range(r_s):
        for y in range(c_s):
            shrunk_img[x, y] = image[x * s:x * s + s, y * s:y * s + s].min()
    return shrunk_img


@jit(nopython=True, cache=True)
def rolling_ball_bg(image, ball):
    width = ball.shape[0]
    radius = int(width / 2)
    r, c = image.shape[0], image.shape[1]
    bg = np.ones((r, c)).astype(np.float32)
    # ignore edges to begin with
    for x in range(r):
        for y in range(c):
            x1, x2, y1, y2 = max(x - radius, 0), min(x + radius + 1, r), max(y - radius, 0), min(y + radius + 1, c)
            cube = image[x1:x2, y1:y2]
            cropped_ball = ball[radius - (x - x1):radius + (x2 - x), radius - (y - y1):radius + (y2 - y)]
            bg_cropped = bg[x1:x2, y1:y2]
            bg_mask = ((cube - cropped_ball).min()) + cropped_ball
            bg[x1:x2, y1:y2] = bg_cropped * (bg_cropped >= bg_mask) + bg_mask * (bg_cropped < bg_mask)
    return (bg)


def subtract_background(data, method='rolling_ball', **kwargs):
    """
    Function to subtract the background from an image.

    :param data: numpy.ndarray
            The input image.
    :param method: str, optional (default='rolling_ball')
            The method used for background subtraction.
            Currently, only 'rolling_ball' is supported.
    :param kwargs: dict
            Additional keyword arguments to be passed to the function.
    :return: numpy.ndarray
            The background-subtracted image.
    :raises ModuleNotFoundError: If the specified method is not found.
    """
    if method == 'rolling_ball':
        return rolling_ball_bg_subtraction(data, **kwargs)
    else:
        raise ModuleNotFoundError('Method {} not found!'.format(method))


"""
=================================== Section 3. Bounding box optimization ===================================
"""

def touching_edge(img_shape, optimized_bbox):
    """
    Check if the bounding box is touching the edge of the image.

    :param img_shape: The shape of the image.
    :param optimized_bbox: The bounding box coordinates.
    :return: True if the bounding box is touching the edge of the image, False otherwise.
    """

    (rows, columns) = img_shape
    (x1, y1, x2, y2) = optimized_bbox
    if min(x1, y1, rows - x2 - 1, columns - y2 - 1) <= 0:
        return True
    else:
        return False


def optimize_bbox(img_shape,
                  bbox,
                  edge_width=8):
    """
    function to optimize bounding box
    :param img_shape: shape of the image to be mapped onto
    :param bbox: inital bbox
    :param edge_width: max edge width
    :return:
    """
    (rows, columns) = img_shape
    (x1, y1, x2, y2) = bbox

    return max(0, x1 - edge_width), max(0, y1 - edge_width), min(rows - 1, x2 + edge_width), min(columns - 1,
                                                                                                 y2 + edge_width)


def optimize_bbox_batch(img_shape,
                        region_prop_table,
                        edge_width=8):
    """
    function to optimize bounding box
    :param img_shape: shape of the image to be mapped onto
    :param bbox: inital bbox
    :param edge_width: max edge width
    :return:
    """

    (rows, columns) = img_shape
    if 'bbox-0' in region_prop_table.columns:
        (x1, y1, x2, y2) = region_prop_table[['bbox-0', 'bbox-1', 'bbox-2', 'bbox-3']].values.T
    elif '$bbox-0' in region_prop_table.columns:
        (x1, y1, x2, y2) = region_prop_table[['$bbox-0', '$bbox-1', '$bbox-2', '$bbox-3']].values.T
    else:
        raise ValueError('Bbox columns should be "bbox-n" or "$bbox-n" where n is 0, 1, 2, or 3.')
    new_x1, new_y1, new_x2, new_y2 = x1 - edge_width, y1 - edge_width, x2 + edge_width, y2 + edge_width
    new_x1[new_x1 <= 0] = 0
    new_y1[new_y1 <= 0] = 0
    new_x2[new_x2 >= rows - 1] = rows - 1
    new_y2[new_y2 >= columns - 1] = columns - 1

    touching_edge = (((x1 <= 0.5 * edge_width) + (y1 <= 0.5 * edge_width) + \
                      (x2 >= rows - 1 - 0.5 * edge_width) + (y2 >= columns - 1 - 0.5 * edge_width)) > 0).astype(int)
    return np.array([new_x1, new_y1, new_x2, new_y2, touching_edge]).T


"""
=================================== Section 4. Bandpass filters ===================================
"""


def fft(img, subtract_mean=True):
    """
    Perform the 2-dimensional Fast Fourier Transform (FFT) on an input image.
    :param img: ndarray, the input image.
    :param subtract_mean: bool, if True the mean of the image is subtracted before transforming, default is True.
    :return: ndarray, the FFT transformed image.
    """
    warnings.filterwarnings("ignore")
    if subtract_mean:
        img = img - np.mean(img)
    return fftpack.fftshift(fftpack.fft2(img))


def fft_reconstruction(fft_img, bandpass_filters):
    """
    Reconstruct an image after applying FFT bandpass filtering.

    :param fft_img: ndarray, the FFT transformed image.
    :param bandpass_filters: list, the low/high frequency bandpass filters.
    :return: ndarray, the bandpass filtered, restored phase contrast image.
    """

    warnings.filterwarnings("ignore")
    if len(bandpass_filters) > 0:
        for f in bandpass_filters:
            try:
                fft_img *= f
            except:
                raise ValueError("Illegal input filter found, shape doesn't match?")
    output = fftpack.ifft2(fftpack.ifftshift(fft_img)).real
    output -= output.min()
    return output


def make_centered_mesh(pixel_microns=0.065,
                       img_width=2048,
                       img_height=2048):
    """
    Create a centered meshgrid in the frequency domain.
    modified to make generic, 20210821
    :param pixel_microns: float, the unit length of a pixel.
    :param img_width: int, the width of the image in pixels.
    :param img_height: int, the height of the image in pixels.
    :return: ndarray, the centered mesh.
    """

    # create mesh
    u_max = round(1 / pixel_microns, 3) / 2
    v_max = round(1 / pixel_microns, 3) / 2
    u_axis_vec = np.linspace(-u_max / 2, u_max / 2, img_width)
    v_axis_vec = np.linspace(-v_max / 2, v_max / 2, img_height)
    u_mat, v_mat = np.meshgrid(u_axis_vec, v_axis_vec)
    centered_mesh = np.sqrt(u_mat ** 2 + v_mat ** 2)
    return centered_mesh


def bandpass_filter(centered_mesh, bandpass_width, highpass=True):
    """
    Create a discrete bandpass filter.

    :param centered_mesh: ndarray, the centered meshgrid.
    :param bandpass_width: float, the width of the bandpass.
    :param highpass: bool, if True a highpass filter is created, otherwise a lowpass filter is created.
    :return: ndarray, the bandpass filter.
    """
    shape = centered_mesh.shape
    if bandpass_width == 0:
        return np.ones((shape[1], shape[0]))
    elif highpass:
        return np.e ** (-(centered_mesh * bandpass_width) ** 2)
    else:
        return 1 - np.e ** (-(centered_mesh * bandpass_width) ** 2)


def dual_bandpass(img,
                  pixel_microns=0.065,
                  min_structure_scale=0.2,
                  max_structure_scale=20):
    """
    Apply a dual-bandpass filter to an image.

    :param img: ndarray, the input image.
    :param pixel_microns: float, the unit length of a pixel.
    :param min_structure_scale: float, the minimum structure size (in microns) for the highpass filter.
    :param max_structure_scale: float, the maximum structure size (in microns) for the lowpass filter.
    :return: ndarray, the image after applying the dual-bandpass filter.

    :raises ValueError: if the min_structure_scale is greater or equal to max_structure_scale.
    """

    if min_structure_scale >= max_structure_scale:
        raise ValueError('Minimal structure size (highpass) should be smaller \nthan maximal structure size (lowpass)')

    fft_img = fft(img)
    img_height, img_width = img.shape
    freq_domain = make_centered_mesh(pixel_microns,
                                     img_width=img_width,
                                     img_height=img_height)
    highpass_filter = bandpass_filter(freq_domain, min_structure_scale, True)
    lowpass_filter = bandpass_filter(freq_domain, max_structure_scale, False)

    # dual-bandpass filter
    filtered = fft_reconstruction(fft_img, (highpass_filter, lowpass_filter))
    return filtered.astype(img.dtype)


"""
=================================== Section 5. Future functions ===================================
"""



